#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-06-30 14:13:45
@File: swanboard/db/migrate.py
@IDE: vscode
@Description:
    可能需要对之前的版本进行兼容
"""
from peewee import CharField
from playhouse.migrate import migrate, SqliteMigrator
from swanboard.db import Tag
from urllib.parse import quote


def compat_tag_key(db):
    """
    对于windows，可能会因为一些保留字的原因而导致tag目录无法创建
    为了以通用方式解决该问题，在0.1.2b8版本以后进行兼容操作：
    为 tag 表新增 folder 字段，该字段对应于实验目录下的 tag 目录
    - 旧版本中，将 tag name 写入 folder 字段
    - 新版本，以 tag id 为 folder 字段值
    """

    migrator = SqliteMigrator(db)
    folder = CharField(default="")
    migrate(migrator.add_column("tag", "folder", folder))

    with db.atomic():
        for tag in Tag.select():
            tag.folder = quote(tag.name, safe="")
            tag.save()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2023-12-02 13:23:42
@File: swanlab\utils\file.py
@IDE: vscode
@Description:
    文件操作
"""
import os
import re
import ujson
import yaml

# ---------------------------------- 一些格式检查的工具函数 ----------------------------------


def check_string(target: str) -> bool:
    """
    检查是否为字符串，且不能全空格，也不能为空字符串
    :param target: 待检查的字符串
    :return: bool
    :raises:
        :raise TypeError: name不是字符串
    """
    if not isinstance(target, str):
        raise TypeError(f"name: {target} is not a string: {type(target)}")
    # 利用正则表达式匹配非空格字符
    if re.match(r"^\s*$", target):
        return False
    # 利用正则表达式匹配非空字符串
    if re.match(r"^\s*$", target) or target == "":
        return False
    return True


# ---------------------------------- 实验、项目相关 ----------------------------------


def _auto_cut(name: str, value: str, max_len: int, cut: bool) -> str:
    """
    检查长度
    :param name: 参数名称
    :param value: 参数值
    :param max_len: 最大长度
    :return: str 检查后的字符串
    :raises
        :raise IndexError: cut为False且name超出长度
    """
    if len(value) > max_len:
        if cut:
            value = value[:max_len]
        else:
            raise IndexError(f"Name: {name} is too long, which must be less than {max_len} characters")
    return value


def check_exp_name_format(name: str, auto_cut: bool = True) -> str:
    """
    检查实验名称格式，最大长度为100个字符，一个中文字符算一个字符
    其他不做限制，实验名称可以包含任何字符

    Parameters
    ----------
    name : str
        待检查的字符串
    auto_cut : bool, optional
        如果超出长度，是否自动截断，默认为True
        如果为False，则超出长度会抛出异常

    Returns
    -------
    str
        检查后的字符串

    Raises
    ------
    TypeError
        name不是字符串
    ValueError
        name为空字符串
    IndexError
        name超出长度
    """
    max_len = 100
    if not check_string(name):
        raise ValueError(f"Experiment name `{name}` is an empty string.")
    # 检查长度
    return _auto_cut("experiment", name, max_len, auto_cut)


def check_desc_format(description: str, auto_cut: bool = True):
    """检查实验描述
    不能超过255个字符，可以包含任何字符

    Parameters
    ----------
    description : str
        需要检查和处理的描述信息
    auto_cut : bool
        如果超出长度，是否裁剪并抛弃多余部分

    Returns
    -------
    str
        检查后的字符串，同时会去除字符串头尾的空格

    Raises
    ------
    IndexError
        name超出长度
    """
    max_length = 255
    check_string(description)
    description = description.strip()
    return _auto_cut("description", description, max_length, auto_cut)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-02-03 16:00:13
@File: swanlab\db\utils\chart.py
@IDE: vscode
@Description:
    关于表格的一些特殊处理
    这里的函数可能在 server、data 中都有作用
"""

from peewee import fn
from ..models import (
    Project,
    Namespace,
    Tag,
    Source,
    Chart,
    Display,
)

from ..db_connect import connect
from ..error import (
    NotExistedError,
    ChartTypeError,
)


def transform_to_multi_exp_charts(project_id: int):
    """兼容以前没有多实验对比数据的情况
    根据早期版本的数据创建多实验对比图表

    Parameters
    ----------
    project_id : int
        project 表 id

    Raises
    ------
    NotExistedError
        project 不存在
    IndexError
        project 已经生成过多实验图表
    NotExistedError
        没有满足多实验图表的 tag
    """

    project = Project.filter(Project.id == project_id)
    if project.count() == 0:
        raise NotExistedError("Target project does not exist")

    if project.first().charts == 1:
        raise IndexError("project.charts must be 0, which means no multiple charts")

    # ---------------------------------- 未生成多实验图表 ----------------------------------

    # 1. 查找
    # - 找到所有满足条件的 tag : tag 名在多个实验中出现
    # - tag 名一样，但类型不一样，以第一个被找到的 tag 类型为基准
    # 2. 生成
    # - 生成 namespace
    # - 生成 chart
    # - 添加 chart 和 namespace 到 display
    # - 生成 source
    # 3. 标记
    # - 设置 project 中字段 charts 为 1，表示已经生成多实验表格

    # 根据名称和类型分类获取 tag
    tags_with_same_name = Tag.select(Tag.name, Tag.type, fn.COUNT(Tag.id).alias("count")).group_by(Tag.name, Tag.type)
    tags = [{"name": tag.name, "type": tag.type} for tag in tags_with_same_name]
    # 如果没有满足多实验图表生成条件的 tag，抛出异常
    # 这是一个非预期异常，因为按道理不会出现这种情况,这种情况也不会写到文档中，仅仅是调试方便
    if len(tags) == 0:
        raise ValueError("No mutiple experment charts found")
    # 用于收集满足多实验图表生成条件的 tag
    result_lists = {}
    # 遍历查询，获取满足多实验图表生成条件的 tag
    for tag in tags:
        # 如果当前 tag 名已经存在收集列表，说明有同名不同类的 tag 都满足多图表条件，目前只取第一次出现的 type，即后续有同名不同类的 tag 都忽略
        if tag["name"] in result_lists and not tag["type"] == Tag.filter(Tag.name == tag["name"]).first().type:
            continue
        # 获取当前name的所有数据行
        target = Tag.filter(Tag.name == tag["name"], Tag.type == tag["type"])
        # 将数据行添加到结果列表
        result_lists[tag["name"]] = [
            {
                "experiment_id": item["experiment_id"]["id"],
                "experiment_name": item["experiment_id"]["name"],
                "tag_id": item["id"],
            }
            for item in Project.search2list(target)
        ]
    # 至此 result_list 数据结构为：{ tag_name: [{experiment_id, experiment_name, tag_id}] }，找到了所有可以生成多实验图表的 tag
    # 查找过滤后，需要生成多实验图表，生成操作需要使用原子操作
    db = connect()
    with db.atomic():
        for tag_name in result_lists:
            tag_list = result_lists[tag_name]
            # 1. 获取/生成 namespace
            # 获取 tag 对应的伴生 namespace 名
            accompanying_chart = Chart.filter(
                Chart.name == tag_name, Chart.experiment_id == tag_list[0]["experiment_id"], Chart.system != 0
            ).first()
            ns_name = Display.filter(Display.chart_id == accompanying_chart.id).first().namespace_id.name
            # 如果属于项目的同名 namespace 已经存在，直接使用这个
            namespaces = Namespace.filter(Namespace.name == ns_name, Namespace.project_id == project_id)
            if namespaces.count():
                namespace = namespaces.first()
            # 没有该命名空间，需创建
            else:
                namespace = Namespace.create(ns_name, project_id=project_id)
            # 2. 生成图表
            chart_type = Source.filter(Source.tag_id == tag_list[0]["tag_id"]).first().chart_id.type
            chart = Chart.create(
                tag_name,
                experiment_id=None,
                project_id=project_id,
                system=0,
                type=chart_type,
            )
            # 3. 添加 chart 和 namespace 到 display
            Display.create(chart_id=chart.id, namespace_id=namespace.id)
            # 4. 生成source
            sources = [
                {
                    "tag_id": item["tag_id"],
                    "chart_id": chart.id,
                    "sort": index,
                    "error": Source.filter(Source.tag_id == item["tag_id"]).first().error,
                }
                for index, item in enumerate(tag_list)
            ]
            Source.insert_many(sources).execute()
        # 最后将 project 表中的 charts 标注为 1，即已经生成多实验表格
        Project.update(charts=1).where(Project.id == project_id).execute()
    db.commit()


def add_multi_chart(
    tag_id: int,
    chart_id: int,
    project_id: int = None,
):
    """添加新的多实验对比图表
    1. 当前 tag 已有多实验图表，将该 tag 添加 source 即可
    2. 当前 tag 满足条件但未创建多实验图表，进行创建和添加

    Parameters
    ----------
    tag_id : int
        tag id
    chart_id : int
        tag 伴生 chart 的 id
    project_id : int
        项目 id, 由于目前为单项目，所以可以不传
    Raises
    ------
    ChartTypeError
        当前 tag 类型和期望 chart 类型不一致
    IndexError
        project 已经生成过多实验图表, 无法再次生成
    """
    if project_id is None:
        project_id = Project.DEFAULT_PROJECT_ID

    try:
        project = Project.filter(Project.id == project_id).first()
    except Exception:
        raise NotExistedError("Target project does not exist")

    # ---------------------------------- charts 为 0 说明还没有生成过多实验图表 ----------------------------------

    if project.charts == 0:
        return transform_to_multi_exp_charts(project_id)

    # ---------------------------------- charts 为 1 说明已经生成过多实验图表 ----------------------------------
    # 当前新增 tag
    tag = Tag.get_by_id(tag_id)
    # 属于项目的所有与 tag 同名的 chart，用于判断是否已经生成该 tag 的多实验对比图
    charts = Chart.filter(Chart.project_id == project_id, Chart.name == tag.name)
    chart = charts.first()

    # 已经有对应名称的 chart
    if charts.count() != 0:
        # 获取图所属 tag 的类型
        if chart.sources.first().tag_id.type != tag.type:
            # 同名 chart 已生成，但是类型不满足
            raise ChartTypeError("Error tag type")
        else:
            # 如果该 chart 和 tag 同类，直接添加
            return Source.create(tag_id, chart.id, error=Source.filter(Source.tag_id == tag_id).first().error)

    # 如果没有对应名称的 chart, 执行创建新多实验对比图表操作
    # 此时该 tag 应该是第一次出现
    # 获取/创建 namespace
    # - 生成或者查找 namespace
    # - 生成 chart
    # - 添加 chart 和 namespace 到 display
    # - 生成 source

    db = connect()
    with db.atomic():
        # 1. 获取 namespace
        name = Display.filter(Display.chart_id == chart_id).first().namespace_id.name
        namespaces = Namespace.filter(Namespace.project_id == project_id, Namespace.name == name)
        # 如果属于项目的同名 namespace 已经存在，直接使用这个
        if Namespace.filter(Namespace.project_id == project_id, Namespace.name == name).count() > 0:
            namespace: Namespace = namespaces.first()
        # 没有该命名空间，需创建, 生成 namespace
        else:
            namespace: Namespace = Namespace.create(name, project_id=project_id)
        # 2. 生成图表
        accompanying_chart: Chart = Chart.get_by_id(chart_id)
        # 通过 tag 的伴生图表获取部分信息
        chart = Chart.create(
            name=accompanying_chart.name,
            project_id=project_id,
            # 不是创建tag时自动生成的图表
            system=0,
            type=accompanying_chart.type,
            reference=accompanying_chart.reference,
        )
        # 3. 添加 chart 和 namespace 到 display
        Display.create(chart.id, namespace.id)
        # 4. 生成 source
        Source.create(
            tag_id,
            chart.id,
            error=Source.filter(Source.tag_id == tag_id).first().error,
        )
    db.commit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-03-10 13:53:16
@File: swanlab/server/controller/chart.py
@IDE: vscode
@Description:
    图表相关操作api
"""
from .db import Chart
from .utils import get_exp_charts, get_proj_charts
from ..module import SUCCESS_200, PARAMS_ERROR_422


def update_charts_status(chart_id: int, status: int):
    """
    更新图表状态，status=1表示置顶，status=0表示正常状态，status=-1表示隐藏

    Parameters
    ----------
    chart_id : int
        图表id
    status : int
        状态，1表示置顶，0表示正常状态，-1表示隐藏
    """
    chart: Chart = Chart.get_by_id(chart_id)

    if not chart:
        return PARAMS_ERROR_422("chart_id not exist")
    if status == 1:
        Chart.pin(id=chart_id)
    elif status == -1:
        Chart.hide(id=chart_id)
    else:
        Chart.restore(id=chart_id)
    if chart.project_id:
        chart_list, namespace_list = get_proj_charts(chart.project_id.id)
    else:
        chart_list, namespace_list = get_exp_charts(chart.experiment_id.id)
    for namespace in namespace_list:
        for j, chart_id in enumerate(namespace["charts"]):
            # 在charts中找到对应的chart_id
            namespace["charts"][j] = next((x for x in chart_list if x["id"] == chart_id), None)
    return SUCCESS_200({"groups": namespace_list})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2023-11-30 20:47:18
@File: swanlab\server\route.py
@IDE: vscode
@Description:
    综合服务 api
"""
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from .middleware.common import (
    resp_base,
    resp_static,
    catch_error,
    log_print,
    resp_params,
)

# 响应路径
from .settings import ASSETS

# 服务全局对象
app = FastAPI()

# 注册前端静态文件路径
static_path = "/assets"
static = StaticFiles(directory=ASSETS)
app.mount(static_path, static)

# 将uvicorn的日志输出handler删除
import logging

# 删除 uvicorn logger
uvicorn_error = logging.getLogger("uvicorn.error")
uvicorn_error.disabled = True
uvicorn_access = logging.getLogger("uvicorn.access")
uvicorn_access.disabled = True


# ---------------------------------- 在此处注册中间件 ----------------------------------


@app.middleware("http")
async def _(*args, **kwargs):
    """基础中间件，调整响应结果，添加处理时间等信息"""
    return await resp_base(*args, **kwargs)


@app.middleware("http")
async def _(*args, **kwargs):
    """资源中间件，此时所有与api相关的内容不会在此中间件中处理"""
    return await resp_static(*args, **kwargs)


@app.middleware("http")
async def _(*args, **kwargs):
    """异常中间件，捕获异常，重构异常信息"""
    return await catch_error(*args, **kwargs)


@app.middleware("http")
async def _(*args, **kwargs):
    """日志打印中间件"""
    return await log_print(*args, **kwargs)


@app.middleware("http")
async def _(*args, **kwargs):
    """参数中间件，处理api请求中的参数校验问题，重新结构化校验错误结果

    参数校验错误并不会影响其他情况的响应结果
    此外由于参数校验错误在绝大多数情况应该是开发时的错误
    所以不会影响正式版本的性能
    """
    return await resp_params(*args, **kwargs)


# ---------------------------------- 在此处注册相关路由 ----------------------------------

# 导入数据相关的路由
from .router.experiment import router as experiment
from .router.project import router as project
from .router.namespace import router as namespace
from .router.chart import router as chart

# 媒体文件路由，允许前端获取其他产生的媒体文件
from .router.media import router as media

# 使用配置列表，统一导入
prefix = "/api/v1"
app.include_router(project, prefix=prefix + "/project")
app.include_router(experiment, prefix=prefix + "/experiment")
app.include_router(media, prefix=prefix + "/media")
app.include_router(namespace, prefix=prefix + "/namespace")
app.include_router(chart, prefix=prefix + "/chart")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-18 15:12:37
@File: swanlab\db\models\namespace.py
@IDE: vscode
@Description:
    实验表格命名空间：一个实验/项目下可以有多个命名空间
"""
from ..model import SwanModel
from peewee import CharField, IntegerField, ForeignKeyField, TextField, IntegrityError, Check, fn, DatabaseProxy
from ..error import ExistedError, ForeignProNotExistedError, ForeignExpNotExistedError
from .experiments import Experiment
from .projects import Project


class Namespace(SwanModel):
    """命名空间表

    Parameters
    ----------
    SwanModel : Class
        自建基类
    """

    class Meta:
        database = DatabaseProxy()

        """
        name 和 experiment_id 和 project_id 组合后需要确保唯一性
        """

        # 同一个项目/实验下，命名空间名称不能重复
        indexes = ((("name", "experiment_id"), True), (("name", "project_id"), True))

        constraints = [
            # project_id 和 experiment_id 不能同时为空，也不能同时不为空
            Check(
                "(project_id IS NULL AND experiment_id IS NOT NULL)"
                " OR "
                "(project_id IS NOT NULL AND experiment_id IS NULL)"
            ),
            # index必须大于等于0
            Check("sort >= 0"),
        ]

    id = IntegerField(primary_key=True)
    """namespace唯一id"""
    experiment_id = ForeignKeyField(
        Experiment, backref="namespaces", on_delete="CASCADE", on_update="CASCADE", null=True
    )
    """外键，对应的实验id，与project_id只能有一个为null"""
    project_id = ForeignKeyField(Project, backref="namespaces", on_delete="CASCADE", on_update="CASCADE", null=True)
    """外键，对应的项目id，与experiment_id只能有一个为null"""
    name = CharField(max_length=100, null=False)
    """这个命名空间的名称，同一个项目/实验下，命名空间名称不能重复"""
    description = CharField(max_length=255, null=True)
    """命名空间描述，可为空"""
    sort = IntegerField()
    """命名空间索引，用于排序，同一个项目/实验下，命名空间索引不能重复，索引越小，排序越靠前，索引>=0"""
    opened = IntegerField(default=1, choices=[0, 1])
    """命名空间是否已经被打开，默认为1，表示已经被打开，0表示未被打开"""
    more = TextField(default=None, null=True)
    """更多信息配置，json格式，将在表函数中检查并解析"""
    create_time = CharField(max_length=30, null=False)
    """创建时间"""
    update_time = CharField(max_length=30, null=False)
    """更新的时间"""

    def __dict__(self):
        return {
            "id": self.id,
            "experiment_id": self.experiment_id.__dict__(),
            "project_id": self.project_id,
            "name": self.name,
            "description": self.description,
            "sort": self.sort,
            "more": self.more,
            "create_time": self.create_time,
            "update_time": self.update_time,
        }

    @classmethod
    def create(
        cls,
        name: str,
        experiment_id: int = None,
        project_id: int = None,
        description: str = None,
        sort=None,
        more: dict = None,
    ) -> "Namespace":
        """创建命名空间

        Parameters
        ----------
        name : str
            命名空间名称
        experiment_id : int, optional
            实验id, 默认为None，但是实验id和项目id不能同时为None，也不能同时不为None
        project_id : int, optional
            项目id, 默认为None，但是实验id和项目id不能同时为None，也不能同时不为None
        description : str, optional
            命名空间描述, 默认为""
        sort : int, optional
            命名空间索引, >=0, 越小排在越前面，默认为None，自动添加到最后
            如果指定了sort，那么将会把其他sort大于等于指定值的命名空间索引+1

        Returns
        -------
        Namespace : Namespace
            创建的命名空间

        Raises
        ------
        ForeignProNotExistedError
            外键对应的id不存在(project不存在)
        ForeignExpNotExistedError
            外键对应的id不存在(experiment不存在)
        ExistedError
           同名命名空间不可存在于同一个实验/项目中
        """

        # 检查外键存在性
        if experiment_id and not Experiment.filter(Experiment.id == experiment_id).exists():
            raise ForeignExpNotExistedError("实验不存在")
        if project_id and not Project.filter(Project.id == project_id).exists():
            raise ForeignProNotExistedError("项目不存在")

        if sort is None:
            # 如果没有指定sort，那么就自动添加到最后
            # 寻找当前实验/项目下的最大索引，如果实验存在就在实验下找，如果实验不存在就在项目下找
            if experiment_id is not None:
                sort = cls.select(fn.Max(cls.sort)).where(cls.experiment_id == experiment_id).scalar()
            else:
                sort = cls.select(fn.Max(cls.sort)).where(cls.project_id == project_id).scalar()
            sort = 0 if sort is None else sort + 1
        elif sort >= 0:
            # 如果指定了sort，那么将会把其他sort大于等于指定值的命名空间索引+1
            if experiment_id is not None:

                cls.update(sort=cls.sort + 1).where(
                    cls.experiment_id == experiment_id,
                    cls.name != name,
                    cls.sort >= sort,
                ).execute()
            else:
                cls.update(sort=cls.sort + 1).where(
                    cls.project_id == project_id,
                    cls.name != name,
                    cls.sort >= sort,
                ).execute()

        try:
            return super().create(
                name=name,
                experiment_id=experiment_id,
                project_id=project_id,
                description=description,
                sort=sort,
                more=more,
            )
        except IntegrityError:
            raise ExistedError("命名空间已存在")



from tutils import (
    mock_project,
    mock_experiment,
    mock_folder_with_run_id,
    mock_experiment_with_folder,
)
import os
import nanoid


def test_mock_project():
    """在数据库中模拟一个项目记录"""
    proj_id = mock_project()
    assert proj_id == 1


def test_mock_experiment():
    """在数据库中模拟创建实验记录"""
    mock_project()
    exp1_id = mock_experiment(1, name="exp1")
    assert exp1_id == 1
    exp2_id = mock_experiment(1, name="exp2")
    assert exp2_id == 2


def test_mock_folder_with_run_id():
    """在测试目录下创建目录"""
    path = mock_folder_with_run_id(nanoid.generate(size=10))
    assert os.path.exists(path)


def test_mock_experiment_with_folder(project_id=1):
    """创建实验记录和对应的运行目录"""
    project_id = mock_project()
    exp_id, path = mock_experiment_with_folder(project_id)
    assert exp_id == 1
    assert os.path.exists(path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2023-12-11 21:41:08
@File: swanlab/utils/color.py
@IDE: vscode
@Description:
    颜色处理工具
"""

import sys
import re
from typing import Callable, Tuple
import threading
import time

light_colors = [
    "#528d59",  # 绿色
    "#587ad2",  # 蓝色
    "#c24d46",  # 红色
    "#9cbe5d",  # 青绿色
    "#6ebad3",  # 天蓝色
    "#dfb142",  # 橙色
    "#6d4ba4",  # 紫色
    "#8cc5b7",  # 淡青绿色
    "#892d58",  # 紫红色
    "#40877c",  # 深青绿色
    "#d0703c",  # 深橙色
    "#d47694",  # 粉红色
    "#e3b292",  # 淡橙色
    "#b15fbb",  # 浅紫红色
    "#905f4a",  # 棕色
    "#989fa3",  # 灰色
]

COLOR_LIST = {
    "light": light_colors,
    "dark": light_colors,
}


def generate_color(number: int = 1):
    """输入数字，在设定好顺序的颜色列表中返回十六进制颜色字符串

    Returns
    -------
    str
        颜色字符串,以#开头的十六进制字符串,如#FFFFFF
        字符串字母大写
    """

    if number % 16 == 0:
        number = 16
    else:
        number = number % 16

    return COLOR_LIST["light"][number - 1], COLOR_LIST["dark"][number - 1]


class FONT:

    @staticmethod
    def loading(
        s: str, func: Callable, args: Tuple = (), interval: float = 0.4, prefix: str = None, brush_length: int = 100
    ):
        """
        实现终端打印的加载效果，输入的字符串会在开头出现loading效果以等待传入的函数执行完毕

        Parameters
        ----------
        s : str
            需要打印的字符串
        func : coroutine
            执行的同步函数
        args : Tuple, optional
            传入函数的参数，默认为空
        interval : float, optional
            loading的速度，即每个字符的间隔时间，单位为秒
        prefix : str, optional
            前缀字符串，打印在loading效果之前，默认为swanlab
        brush_length : int, optional
            刷去的长度，默认为100
        """
        # FIXME 因为协程有一些适配性问题，暂时使用线程
        prefix = FONT.bold(FONT.blue("swanlab")) + ": " if prefix is None else prefix
        symbols = ["\\", "|", "/", "-"]

        running, result, error = True, None, None

        def loading():
            index = 0
            while True:
                sys.stdout.write("\r" + prefix + symbols[index % len(symbols)] + " " + s)
                sys.stdout.flush()
                index += 1
                time.sleep(interval)
                if not running:
                    break

        # 再次封装传入的func，当func执行完毕以后，将running置为False
        def task():
            nonlocal result, error, running
            try:
                result = func(*args)
            except Exception as e:
                error = e
            finally:
                running = False

        # 开启新线程
        t1 = threading.Thread(target=loading)
        t2 = threading.Thread(target=task)
        t1.start()
        t2.start()
        try:
            t2.join()
        except KeyboardInterrupt:
            running = False
            raise KeyboardInterrupt
        t1.join()
        if error is not None:
            raise error
        FONT.brush("", brush_length)
        return result

    @staticmethod
    def swanlab(s: str, color: str = "blue"):
        """用于为某一条信息添加swanlab前缀"""
        return FONT.bold(getattr(FONT, color)("swanlab")) + ": " + s

    @staticmethod
    def brush(s: str, length: int = 20) -> None:
        """
        将当前终端行刷去，替换为新的字符串

        Parameters
        ----------
        s : str
            需要刷去的字符串
        length : int, optional
            需要刷去的长度，默认为20，如果当前行的长度大于length，但又需要刷去整行，则需要传入更大的length
        """
        sys.stdout.write("\r" + " " * length + "\r" + s)
        sys.stdout.flush()

    @staticmethod
    def bold(s: str) -> str:
        """在终端中加粗字符串

        Parameters
        ----------
        s : str
            需要加粗的字符串

        Returns
        -------
        str
            加粗后的字符串
        """
        # ANSI 转义码用于在终端中改变文本样式
        return f"\033[1m{s}\033[0m"

    @staticmethod
    def default(s: str) -> str:
        """在终端中将字符串着色为默认颜色

        Parameters
        ----------
        s : str
            需要着色的字符串

        Returns
        -------
        str
            着色后的字符串
        """
        # ANSI 转义码用于在终端中改变文本样式
        return f"\033[39m{s}\033[0m"

    @staticmethod
    def blue(s: str) -> str:
        """在终端中将字符串着色为蓝色

        Parameters
        ----------
        s : str
            需要着色的字符串

        Returns
        -------
        str
            着色后的字符串
        """
        # ANSI 转义码用于在终端中改变文本样式
        return f"\033[34m{s}\033[0m"

    @staticmethod
    def grey(s: str) -> str:
        """在终端中将字符串着色为灰色

        Parameters
        ----------
        s : str
            需要着色的字符串

        Returns
        -------
        str
            着色后的字符串
        """
        # ANSI 转义码用于在终端中改变文本样式
        return f"\033[90m{s}\033[0m"

    @staticmethod
    def underline(s: str) -> str:
        """在终端中将字符串着色为下划线

        Parameters
        ----------
        s : str
            需要着色的字符串

        Returns
        -------
        str
            着色后的字符串
        """
        # ANSI 转义码用于在终端中改变文本样式
        return f"\033[4m{s}\033[0m"

    @staticmethod
    def green(s: str) -> str:
        """在终端中将字符串着色为绿色

        Parameters
        ----------
        s : str
            需要着色的字符串

        Returns
        -------
        str
            着色后的字符串
        """
        # ANSI 转义码用于在终端中改变文本样式
        return f"\033[32m{s}\033[0m"

    @staticmethod
    def dark_green(s: str) -> str:
        """在终端中将字符串着色为深绿色

        Parameters
        ----------
        s : str
            需要着色的字符串

        Returns
        -------
        str
            着色后的字符串
        """
        # ANSI 转义码用于在终端中改变文本样式
        return f"\033[38;5;22m{s}\033[0m"

    @staticmethod
    def dark_gray(s: str) -> str:
        """在终端中将字符串着色为深灰色

        Parameters
        ----------
        s : str
            需要着色的字符串

        Returns
        -------
        str
            着色后的字符串
        """
        # ANSI 转义码用于在终端中改变文本样式
        return f"\033[38;5;236m{s}\033[0m"

    @staticmethod
    def yellow(s: str) -> str:
        """在终端中将字符串着色为黄色

        Parameters
        ----------
        s : str
            需要着色的字符串

        Returns
        -------
        str
            着色后的字符串
        """
        # ANSI 转义码用于在终端中改变文本样式
        return f"\033[33m{s}\033[0m"

    @staticmethod
    def red(s: str) -> str:
        """在终端中将字符串着色为红色

        Parameters
        ----------
        s : str
            需要着色的字符串

        Returns
        -------
        str
            着色后的字符串
        """
        # ANSI 转义码用于在终端中改变文本样式
        return f"\033[31m{s}\033[0m"

    @staticmethod
    def magenta(s: str) -> str:
        """在终端中将字符串着色为品红色

        Parameters
        ----------
        s : str
            需要着色的字符串

        Returns
        -------
        str
            着色后的字符串
        """
        # ANSI 转义码用于在终端中改变文本样式
        return f"\033[35m{s}\033[0m"

    @staticmethod
    def clear(s: str) -> str:
        """清除字符串中的颜色编码

        Parameters
        ----------
        s : str
            需要清除颜色的字符串

        Returns
        -------
        str
            清除颜色后的字符串
        """
        ansi_escape_pattern = re.compile(r"\033\[[0-9;]+m")
        return ansi_escape_pattern.sub("", s)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-06-26 15:49:02
@File: test/unit/db/pytest_connect.py
@IDE: vscode
@Description:
    测试数据库连接函数
"""

from swanboard.db import connect, tables
from tutils import create_test_dir
import os
import shutil


class TestConnectDB:
    """
    因为在每个单元测试开启之前都会自动在默认目录下连接数据库
    所以若要对数据库的创建/连接情况进行测试，需要重新创建目录
    """

    def test_create_db(self):
        path = create_test_dir("test_for_connect")
        connect(path, autocreate=True)
        # 在path下有 run.swanlab 文件
        assert os.path.exists(os.path.join(path, "runs.swanlab"))
        if os.path.exists(path):
            shutil.rmtree(path)

    def test_create_tables(self):
        """测试是否自动创建数据表"""
        path = create_test_dir("test_for_connect")
        db = connect(path, autocreate=True)
        # 在数据库中有各个数据表
        for table in tables:
            assert db.table_exists(table)
        if os.path.exists(path):
            shutil.rmtree(path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-18 20:36:01
@File: swanlab\db\models\sources.py
@IDE: vscode
@Description:
    中间表，多chart多tag
"""
from .tags import Tag
from .charts import Chart
from ..model import SwanModel
from peewee import CharField, IntegerField, ForeignKeyField, TextField, IntegrityError, DatabaseProxy, fn
from ..error import ExistedError, ForeignTagNotExistedError, ForeignChartNotExistedError


class Source(SwanModel):
    """多chart多tag"""

    class Meta:
        database = DatabaseProxy()
        # 写入check约束，name和experiment_id加起来唯一
        indexes = ((("tag_id", "chart_id"), True),)

    id = IntegerField(primary_key=True)
    """此表唯一id"""
    tag_id = ForeignKeyField(Tag, backref="sources", null=False, on_delete="CASCADE", on_update="CASCADE")
    """代表关联的tag，不可为空"""
    chart_id = ForeignKeyField(Chart, backref="sources", null=False, on_delete="CASCADE", on_update="CASCADE")
    """关联的chart，不可为空"""
    sort = IntegerField(default=0)
    """此tag在此chart中的排序，值越小越靠前"""
    error = TextField(null=True)
    """代表这个tag数据转化到chart上失败了，错误信息，是一个json字符串"""
    more = TextField(null=True)
    """更多信息配置，json格式，将在表函数中检查并解析"""
    create_time = CharField(max_length=30, null=False)
    """创建时间"""
    update_time = CharField(max_length=30, null=False)
    """更新时间"""

    def __dict__(self):
        return {
            "id": self.id,
            "tag_id": self.tag_id,
            "chart_id": self.chart_id,
            "sort": self.sort,
            "error": self.error,
            "more": self.more,
            "create_time": self.create_time,
            "update_time": self.update_time,
        }

    @classmethod
    def create(
        cls,
        tag_id: int,
        chart_id: int,
        error: dict = None,
        sort: int = None,
        more: dict = None,
    ) -> "Source":
        """添加行数据

        Parameters
        ----------
        tag_id : int
            对应的tag_id
        chart_id : int
            对应的chart_id
        error : str, optional
            对应的错误信息, 默认为None
        sort : int, optional
            排序索引，索引越小，排序越靠前，索引>=0，如果为None，则自动加到最后
        more : str, optional
            更多信息，json格式

        Raises
        ------
        ForeignTagNotExistedError
            外键对应的id不存在(tag不存在)
        ForeignChartNotExistedError
            外键对应的id不存在(chart不存在)
        ExistedError
           对应关系已经在数据库中存在（"tag_id"和"chart_id"唯一）
        """

        # 检查外键存在性
        if not Tag.filter(Tag.id == tag_id).exists():
            raise ForeignTagNotExistedError("tag不存在")
        if not Chart.filter(Chart.id == chart_id).exists():
            raise ForeignChartNotExistedError("chart不存在")

        # 如果sort为None，则自动添加到最后
        if sort is None:
            sort = Source.select(fn.Max(Source.sort)).where(Source.chart_id == chart_id).scalar()
            sort = sort + 1 if sort is not None else 0
        else:
            # 如果sort不为None，则检查当前chart下的sort是否存在，如果存在，则把sort大于等于sort的索引+1
            cls.update(sort=sort + 1).where(cls.chart_id == chart_id, cls.sort >= sort).execute()

        try:
            return super().create(
                tag_id=tag_id,
                chart_id=chart_id,
                error=cls.dict2json(error),
                more=cls.dict2json(more),
                sort=sort,
            )
        except IntegrityError:
            raise ExistedError("source对应关系已存在")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-03-17 15:13:31
@File: swanlab/server/controller/utils/tag.py
@IDE: vscode
@Description:
    tag相关处理函数
"""
from typing import List
import json
import ujson
import math
import os

# tag 总结文件名
TAG_SUMMARY_FILE = "_summary.json"
# logs 目录下的配置文件
LOGS_CONFIGS = [TAG_SUMMARY_FILE]


def read_tag_data(file_path: str) -> List[dict]:
    """读取某一个tag文件数据，数据依据其中的index字段排序

    Parameters
    ----------
    file_path : str
        tag文件路径，绝对路径
    """
    # 如果文件内容为空，返回空列表
    if os.path.getsize(file_path) == 0:
        return []
    # 读取文件内容，文件内部本质上是字符串一堆json格式的数据，每一行是一个json并且有换行符分隔，需要拿到然后解析为list<dict>
    data = []
    with open(file_path, "r") as f:
        lines = f.readlines()
        # 解析为list<dict>
        for i in range(len(lines)):
            if len(lines[i]):
                data.append(json.loads(lines[i]))
        return data


def get_tag_files(tag_path: str, exclude=None) -> List[str]:
    """
    获取实验数据

    Parameters
    ----------
    tag_path : str
        tag的路径
    exclude : List[str]
        需要排除的文件列表

    Returns
    -------
    List[str]
        tag文件列表
    """
    # 降序排列，最新的数据在最前面
    if exclude is None:
        exclude = []
    files: list = os.listdir(tag_path)
    current_logs = [f for f in files if f.endswith(".log")]
    current_logs.sort()
    return current_logs


def _sample_a_bucket(bucket: List[dict]) -> dict:
    """
    对一个bucket进行采样，返回采样后的数据
    :param bucket: 一个桶中的信息
    :return: 采样后的数据
    """
    if len(bucket) <= 2:
        return bucket[0]
    # 计算bucket头尾直线表达式
    tail = bucket[0]
    head = bucket[-1]
    # k，b
    k = (head["data"] - tail["data"]) / (head["index"] - tail["index"])
    b = head["data"] - k * head["index"]
    # 计算每个点到直线的距离
    max_distance_p = (0, 0)  # (d, i)
    for i in range(1, len(bucket) - 1):
        distance = abs(k * bucket[i]["index"] + b - bucket[i]["data"]) / math.sqrt(k**2 + 1)
        if distance > max_distance_p[0]:
            max_distance_p = (distance, i)
    return bucket[max_distance_p[1]]


def _calculate_bucket_capacity(total_data_points, target_bucket_count) -> List[int]:
    """
    计算每个桶的容量
    :param total_data_points: 总数据点数
    :param target_bucket_count: 目标桶数量
    :return: 每个桶的容量
    """
    # 计算每个桶的容量
    bucket_capacity = total_data_points // target_bucket_count  # 整除部分
    remaining_points = total_data_points % target_bucket_count  # 余数

    # 将余数均匀地分配到前几个桶中
    bucket_capacity_list = [bucket_capacity] * target_bucket_count
    for i in range(remaining_points):
        bucket_capacity_list[i] += 1

    return bucket_capacity_list


def lttb(data: List[dict], threshold: int = 1500):
    """
    LTTB算法，对数据进行降采样
    最后保留头尾数据，因此桶的数量应该是实际数量-2
    桶大小向下取整，保留第一个和最后一个点
    降采样将优先发生在尾部
    :param data: 数据
    :param threshold: 采样后的数据长度，也是降采样阈值
    """
    if len(data) <= threshold:
        return data
    bucket_n = threshold - 2
    _data = data[1:-1]
    # 需要保留第一、最后一个点
    sampled = [data[0]]
    # 计算每一个桶的容量，每个桶容量不尽相同
    buckets_capacity = _calculate_bucket_capacity(len(_data), bucket_n)
    # 当前使用的bucket
    now_bucket = []
    # 当前bucket的数量
    now_bucket_n = 0
    for _d in _data:
        now_bucket.append(_d)
        if len(now_bucket) < buckets_capacity[now_bucket_n]:
            continue
        sampled.append(_sample_a_bucket(now_bucket))
        now_bucket = []
        now_bucket_n += 1
        if now_bucket_n == bucket_n - 1:
            break
    sampled.append(data[-1])
    return sampled



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-19 22:10:15
@File: swanlab\server\middleware\common.py
@IDE: vscode
@Description:
    常规中间件
"""

from fastapi import Request
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
import time
from ..module.resp import UNEXPECTED_ERROR_500, PARAMS_ERROR_422
from swanboard.utils import swanlog as swl, get_package_version
from ..settings import ASSETS, INDEX  # 响应路径

version = get_package_version()

# 注册静态文件路径
static_path = "/assets"
static = StaticFiles(directory=ASSETS)


async def resp_base(request, call_next):
    """基础中间件，调整响应结果，添加处理时间等信息"""
    # 如果请求路径不以 '/api' 开头，说明并不是后端服务的请求，直接返回
    if not request.url.path.startswith("/api"):
        return await call_next(request)
    # 记录请求开始时间
    start_time = time.time()
    # 调用下一个中间件或者最终的路由处理函数
    # 我们约定路由处理函数最终返回三个参数，一个是错误码，一个是错误信息，一个是响应内容
    response = await call_next(request)
    # 记录请求结束时间
    end_time = time.time()
    # 计算处理时间，添加到响应头中
    process_time = round(end_time - start_time, 4)
    response.headers["SwanLab-Process-Time"] = str(process_time)
    # 添加版本信息
    response.headers["SwanLab-Version"] = version
    # 返回响应
    return response


async def resp_static(request, call_next):
    """资源中间件，此时所有与api相关的内容不会在此中间件中处理"""
    if request.url.path.startswith(static_path):
        # 如果是请求静态资源，直接返回
        return await call_next(request)
    if request.url.path.startswith("/api"):
        # 如果是请求api，直接返回
        return await call_next(request)
    # 剩余情况返回index.html，交由前端路由处理
    with open(INDEX, "r", encoding="utf-8") as file:
        html_content = file.read()
    return HTMLResponse(content=html_content, status_code=200)


async def catch_error(request: Request, call_next):
    """异常中间件，捕获异常，重构异常信息"""
    if not request.url.path.startswith("/api"):
        # 如果不是请求api，直接返回
        return await call_next(request)
    try:
        return await call_next(request)
    except Exception as e:
        swl.error(str(e))
        return UNEXPECTED_ERROR_500(str(e))


async def log_print(request: Request, call_next):
    """日志打印中间件"""
    swl.debug("[" + request.method + "] from " + request.base_url._url)
    resp = await call_next(request)
    # 拿到状态码
    # status = str(resp.status_code)
    if not request.url.path.startswith("/api"):
        # 如果不是请求api，直接返回
        swl.debug("[" + str(resp.status_code) + "] " + request.method + " assets: " + request.url.path)
    else:
        content = "[" + str(resp.status_code) + "] " + request.method + " api: " + request.url.path
        swl.debug(content)
    return resp


async def resp_params(request: Request, call_next):
    """参数中间件，处理api请求中的参数校验问题，重新结构化校验错误结果"""
    if not request.url.path.startswith("/api"):
        # 如果不是请求api，直接返回
        return await call_next(request)
    resp = await call_next(request)
    # 拿到状态码
    status = resp.status_code
    if status == 422:
        # 参数校验错误，重构错误信息
        # 拿到响应体
        import json

        body = [chunk async for chunk in resp.body_iterator][0].decode()
        body = json.loads(body)
        detail = body["detail"][0]
        msg = detail["msg"].split(" ")[1:]
        loc = detail["loc"][1]
        msg.insert(0, loc)
        msg = " ".join(msg)
        return PARAMS_ERROR_422(msg)
    """
    参数校验错误并不会影响其他情况的响应结果
    此外由于参数校验错误在绝大多数情况应该是开发时的错误
    所以不会影响正式版本的性能
    """
    return resp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-18 13:08:48
@File: swanlab/db/models/__init__.py
@IDE: vscode
@Description:
    在此处自定义模型类，继承自SwanModel
"""
from .charts import Chart
from .displays import Display
from .experiments import Experiment
from .namespaces import Namespace
from .projects import Project
from .tags import Tag
from .sources import Source



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-03-09 21:41:37
@File: swanlab/server/controller/utils/__init__.py
@IDE: vscode
@Description:
    共享工具函数
"""
from .charts import get_exp_charts, get_proj_charts
from .tag import read_tag_data, get_tag_files, LOGS_CONFIGS, lttb
from typing import List


def clear_field(target: List[dict], field: str) -> List[dict]:
    """遍历字典列表清除某个字段

    Parameters
    ----------
    target : List[dict]
        需要处理的列表
    field : str
        需要删除的字段

    Returns
    -------
    List[dict]
        处理后的字典列表
    """

    for item in target:
        item.pop(field)

    return target



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-18 13:11:59
@File: swanlab/db/model.py
@IDE: vscode
@Description:
    在此处定义基础模型类
"""
from peewee import Model, OperationalError
from playhouse.shortcuts import model_to_dict
from swankit.env import create_time
from .error import NotExistedError
import json


class SwanModel(Model):
    """基础模型类，用于定义数据库表的基本信息"""

    @staticmethod
    def search2dict(result) -> dict:
        """将select、filter的结果转换为字典"""

        dicts = [model_to_dict(row) for row in result]
        if len(dicts) == 1:
            return dicts[0]
        else:
            return dicts

    @staticmethod
    def search2list(result) -> list:
        """将select、filter的结果转换为列表"""

        return [model_to_dict(row) for row in result]

    @staticmethod
    def json2dict(json_str: str) -> dict:
        """将json字符串转换为字典
        如果字符串为空，返回空字典

        Parameters
        ----------
        json_str : str
            传入的json格式字符串，或者空字符串

        Returns
        ----------
        dict :
            转换后的字典

        Raises
        ----------
        TypeError :
            无法转换为字典
        """
        if json_str == "":
            return {}
        try:
            return json.loads(json_str)
        except:
            raise TypeError

    @staticmethod
    def dict2json(dict: dict) -> str:
        """将字典转换为json字符串,如果已经是json字符串，则直接返回

        Parameters
        ----------
        dict : dict
            输入的字典，如果dict为None，则解析为空字符串

        Returns
        -------
        str
            转换后的字符串

                Raises
        ----------

        TypeError :
            无法转换为Json字符串
        """
        if dict is None:
            return ""
        if isinstance(dict, str):
            try:
                if not dict:
                    return ""
                json.loads(dict)
                return dict
            except:
                raise TypeError
        try:
            return json.dumps(dict)
        except:
            raise TypeError

    def save(self, *args, **kwargs):
        """保存数据
        经过swanmodel的覆写，save方法自动更新update_time字段
        """
        self.update_time = create_time()
        super().save(*args, **kwargs)

    @classmethod
    def insert_many(cls, rows, fields=None):
        """批量插入数据
        经过swanmodel的覆写，insert_many方法自动创建create_time和update_time字段
        """
        current_time = create_time()
        rows = [{**row, "create_time": current_time, "update_time": current_time} for row in rows]
        return super().insert_many(rows, fields=fields)

    @classmethod
    def create(cls, **query):
        current_time = create_time()
        query["create_time"] = current_time
        query["update_time"] = current_time
        return super().create(**query)

    @classmethod
    def field_exists(cls, field: str) -> bool:
        """
        判断某个字段是否存在于表中

        Parameters
        ----------
        field : str
            字段名
        """
        # 进行一次查询
        try:
            a = cls.select().where(cls.__getattribute__(cls, field) is not None)
            cls.search2dict(a)
        except OperationalError:
            return False
        return True



from fastapi import APIRouter
from ..module.resp import PARAMS_ERROR_422

from ..controller.namespace import (
    # 修改namespace开启关闭状态
    change_namespace_opened,
)
from fastapi import Request
from urllib.parse import quote

router = APIRouter()


# 修改namespace可见性
@router.patch("/{namespace_id}/opened")
async def _(namespace_id: int, request: Request):
    """修改namespace可见性

    Parameters
    ----------
    experiment_id : int
        实验id
    request : Request
        请求体，包含 opened 字段
    Returns
    -------
    experiment : dict
        当前实验信息
    """

    data = await request.json()
    opened = data.get("opened")
    experiment_id = data.get("experiment_id")
    project_id = data.get("project_id")

    if opened is None:
        return PARAMS_ERROR_422("Request parameter 'opened'")
    if experiment_id is None and project_id is None:
        return PARAMS_ERROR_422("Request parameter 'experiment_id' or 'project_id'")
    return change_namespace_opened(namespace_id, opened, experiment_id, project_id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-18 20:43:32
@File: swanlab\db\models\displays.py
@IDE: vscode
@Description:
    中间表，namespace 和 chart，设计排序和排序可变
"""
from ..model import SwanModel
from peewee import CharField, IntegerField, ForeignKeyField, TextField, IntegrityError, Check, fn, DatabaseProxy
from ..error import ExistedError, ForeignChartNotExistedError, ForeignNameNotExistedError
from .charts import Chart
from .namespaces import Namespace


class Display(SwanModel):
    """namespace 和 chart，设计排序和排序可变"""

    class Meta:
        database = DatabaseProxy()
        # chart_id和namespace_id加起来唯一
        indexes = ((("chart_id", "namespace_id"), True),)
        # 写入check约束，sort必须大于等于0
        constraints = [Check("sort >= 0")]

    id = IntegerField(primary_key=True)
    """display表唯一id"""
    chart_id = ForeignKeyField(Chart, backref="displays", on_delete="CASCADE", on_update="CASCADE", null=False)
    """关联的chart，不可为空"""
    namespace_id = ForeignKeyField(Namespace, backref="displays", on_delete="CASCADE", on_update="CASCADE", null=False)
    """关联的namespace，不可为空"""
    sort = IntegerField()
    """当前chart在namespace下的排序，索引越小，排序越靠前，索引>=0"""
    more = TextField(default=None, null=True)
    """更多信息配置，json格式，将在表函数中检查并解析"""
    create_time = CharField(max_length=30, null=False)
    """创建时间"""
    update_time = CharField(max_length=30, null=False)
    """更新时间"""

    def __dict__(self):
        return {
            "id": self.id,
            "chart_id": self.chart_id,
            "namespace_id": self.namespace_id,
            "sort": self.sort,
            "more": self.more,
            "create_time": self.create_time,
            "update_time": self.update_time,
        }

    @classmethod
    def create(
        cls,
        chart_id: int,
        namespace_id: int,
        sort: int = None,
        more: dict = None,
    ) -> "Display":
        """添加行数据

        Parameters
        ----------
        chart_id : int
            chart表中的id
        namespace_id : int
            namespace表中的id
        sort : int
            排序索引，索引越小，排序越靠前，索引>=0，如果为None，则自动加到最后
        more : str
            更多信息配置，json格式，将在表函数中检查并解析

        Returns
        -------
        Display
            返回创建的display对象

        Raises
        ------
        ForeignChartNotExistedError
            外键对应的id不存在(chart不存在)
        ForeignNameNotExistedError
            外键对应的id不存在(namespace不存在)
        ExistedError
           "chart_id" 和 "namespace_id" 的对应关系已存在
        """

        # 检查外键存在性
        if not Chart.filter(Chart.id == chart_id).exists():
            raise ForeignChartNotExistedError("图表不存在")
        if not Namespace.filter(Namespace.id == namespace_id).exists():
            raise ForeignNameNotExistedError("命名空间不存在")

        # 如果sort为None，则自动添加到最后
        if sort is None:
            # 获取当前namespace下的最大索引,如果没有，则为0
            sort = cls.select(fn.Max(cls.sort)).where(cls.namespace_id == namespace_id).scalar()
            sort = 0 if sort is None else sort + 1
        elif sort >= 0 and cls.filter(cls.sort == sort).exists():
            # 先判断当前 sort 是否存在，如果存在，则将大于等于 sort 的索引加1，这样可以直接将数据插入
            if cls.filter(cls.sort == sort).exists():
                cls.update(sort=cls.sort + 1).where(cls.sort >= sort).execute()

        try:
            return super().create(
                chart_id=chart_id,
                namespace_id=namespace_id,
                sort=sort,
                more=cls.dict2json(more),
            )
        except IntegrityError:
            raise ExistedError("display对应关系已存在")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-19 22:09:09
@File: swanlab\server\controller\project.py
@IDE: vscode
@Description:
    项目相关 api 的处理函数
"""

import os
import ujson
import shutil
from fastapi import Request
from swanboard.utils import check_desc_format, get_swanlog_dir, COLOR_LIST, swanlog
import yaml
from typing import List
from ..db.utils.chart import transform_to_multi_exp_charts
from .db import NotExistedError
from .utils import get_proj_charts

# 自定义响应
from ..module.resp import (
    SUCCESS_200,
    DATA_ERROR_500,
    CONFLICT_409,
    NOT_FOUND_404,
)

# 功能性函数
from ..settings import (
    get_tag_dir,
    get_exp_dir,
    get_config_path,
)

# 数据表模型
from ..db import (
    Project,
    Experiment,
    Tag,
    Chart,
    Namespace,
)
from .utils import get_tag_files, LOGS_CONFIGS

# 将查询结果对象转为列表
__to_list = Project.search2list

# ---------------------------------- 通用 ----------------------------------

# 默认项目 id
DEFAULT_PROJECT_ID = Project.DEFAULT_PROJECT_ID
# 实验运行状态
RUNNING_STATUS = Experiment.RUNNING_STATUS


# ---------------------------------- 路由对应的处理函数 ----------------------------------


# 获取项目信息
def get_project_info(project_id: int = DEFAULT_PROJECT_ID) -> dict:
    """
    1. 获取项目信息
    2. 列出当前项目下的所有实验
    3. 获取 swanlog 目录路径

    Parameters
    ----------
    project_id : int, optional
        项目id, 默认为DEFAULT_PROJECT_ID

    Returns
    -------
    dict:
        列出当前项目下的所有实验
    """

    try:
        project = Project.filter(Project.id == project_id).first()
        data = project.__dict__()
        data["logdir"] = get_swanlog_dir()
        experiments = __to_list(project.experiments)
        for experiment in experiments:
            experiment["experiment_id"] = experiment["id"]
        data["experiments"] = experiments
        for experiment in data["experiments"]:
            # 将其中的 project_id 字段去除
            experiment.pop("project_id")
            # 检查配置文件是否存在
            config_path = get_config_path(experiment["run_id"])
            if os.path.exists(config_path):
                # 加载config字段
                with open(config_path, "r+", encoding="utf-8") as f:
                    experiment["config"] = yaml.load(f, Loader=yaml.FullLoader)
        # 色盘
        data["colors"] = COLOR_LIST
        return SUCCESS_200(data)
    except Exception as e:
        return DATA_ERROR_500(f"Get list experiments failed: {e}")


# 获取项目总结信息
def get_project_summary(project_id: int = DEFAULT_PROJECT_ID) -> dict:
    """
    获取项目下所有实验的总结信息

    Parameters
    ----------
    project_id : int, optional
        项目id, 默认为DEFAULT_PROJECT_ID

    Returns
    -------
    dict:
        项目总结信息
    """

    # 查找所有实验，提出 id 列表
    experiments = Experiment.select().where(Experiment.project_id == project_id)
    exprs = [
        {
            "id": experiment.id,
            "name": experiment.name,
            "run_id": experiment.run_id,
            "tags": [{"name": tag["name"], "sort": tag["sort"]} for tag in __to_list(experiment.tags)],
        }
        for experiment in experiments
    ]
    ids = [item["id"] for item in exprs]
    # 根据 id 列表找到所有的 tag，提出不含重复 tag 名的元组
    # tags = Tag.filter(Tag.experiment_id.in_(ids)).order_by(Tag.experiment_id)
    tags = (
        Tag.select()
        .join(Experiment, on=(Tag.experiment_id == Experiment.id))
        .where(Tag.experiment_id.in_(ids))
        .order_by(Tag.experiment_id, Tag.create_time)
    )
    # tag_names不用编码，因为前端需要展示
    # 要保持顺序，需要额外使用一个列表，直接用 set 会打乱顺序
    unique_values = []
    seen = set()
    for value in [item["name"] for item in __to_list(tags)]:
        if value not in seen:
            unique_values.append(value)
            seen.add(value)
    tag_names = list(unique_values)
    # 所有总结数据
    data = {}
    # 第一层循环对应实验层，每次探寻一个实验
    for expr in exprs:
        # 第二层为 tag 层，在实验目录下遍历所有 tag，若存在则获取最后一次提交的值
        experiment_summaries = {}
        for tag in expr["tags"]:
            # tag 结构：{ name: tag_name, sort: tag_sort }，其中 sort 同时为 tag 文件夹名
            tag_path = get_tag_dir(expr["run_id"], str(tag["sort"]))
            if not os.path.exists(tag_path):
                experiment_summaries[tag["name"]] = "TypeError"
                continue
            # 获取所有 tag 文件
            logs = get_tag_files(tag_path, exclude=LOGS_CONFIGS)
            # 打开最后一个文件，获取最后一行数据
            with open(os.path.join(tag_path, logs[-1]), mode="r", encoding="utf-8") as f:
                try:
                    lines = f.readlines()
                    tag_data = ujson.loads(lines[-1])
                    experiment_summaries[tag["name"]] = tag_data["data"]
                except Exception as e:
                    swanlog.error(f"[expr: {expr['name']} - {tag['name']}] --- {e}")
                    continue

        data[expr["name"]] = experiment_summaries

    return SUCCESS_200({"tags": tag_names, "summaries": data})


# 修改项目信息
async def update_project_info(request: Request, project_id: int = DEFAULT_PROJECT_ID) -> dict:
    """修改项目信息
    - 项目名
    - 项目描述

    Parameters
    ----------
    project_id : int, optional
        默认为 DEFAULT_PROJECT_ID

    Returns
    -------
    dict
        新项目信息
        - name
        - description
    """

    body = await request.json()

    # 检查格式
    body["description"] = check_desc_format(body["description"], False)

    project = Project.filter(Project.id == project_id).first()
    dict_project = project.__dict__()

    # 检查名字
    if "name" in dict_project and dict_project["name"] == body["name"]:
        pass
    else:
        project.name = body["name"]

    # 检查描述
    if "description" in dict_project and dict_project["description"] == body["description"]:
        pass
    else:
        project.description = body["description"]

    project.save()
    return SUCCESS_200({"updates": body})


# 删除项目
async def delete_project(project_id: int = DEFAULT_PROJECT_ID):
    """删除项目
    1. 删除实验目录
    2. 删除数据库

    Parameters
    ----------
    project_id : int, optional
        默认为第一个项目

    TODO: 原子操作，同时删除项目和实验
    """

    # 检查是否有正在运行的实验
    running_exp = Experiment.filter(Experiment.project_id == project_id, Experiment.status == RUNNING_STATUS).count()
    if running_exp > 0:
        return CONFLICT_409("Can't delete project since there is experiment running")

    project = Project.filter(Project.id == project_id).first()
    run_ids = [experiment["run_id"] for experiment in __to_list(project.experiments)]

    # 删除实验目录
    for run_id in run_ids:
        exp_path = get_exp_dir(run_id)
        if os.path.isdir(exp_path):
            shutil.rmtree(exp_path)

    # 清空数据库
    project.delete_instance()

    return SUCCESS_200({})


async def get_project_charts(project_id: int = DEFAULT_PROJECT_ID) -> dict:
    """获取多实验对比图表数据,并且考虑往期版本兼容性
    1. 如果当前项目的chart字段为0，先生成多实验对比数据，跳转步骤2
    2. 依据规则获取所有实验的图表数据
    """

    # COMPAT 兼容以前没有多实验对比数据的情况
    try:
        transform_to_multi_exp_charts(project_id)
    except IndexError:  # 已经有多实验对比数据
        pass
    charts, namespace = get_proj_charts(project_id)
    # 获取项目下所有实验的图表数据
    return SUCCESS_200({"charts": charts, "namespaces": namespace})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-02-05 16:56:06
@File: swanlab/db/connect.py
@IDE: vscode
@Description:
    数据库连接模块
"""
import os
from peewee import SqliteDatabase
from .table_config import tables, Tag
from .migrate import compat_tag_key

db_path = None
"""
全局挂载的数据库路径，当autocreate设置为True时，必须指定此路路径
可用于判断数据库是否已经创建
"""


def connect(path: str = None, autocreate: bool = False) -> SqliteDatabase:
    """
    连接数据库，只有调用此方法以后，数据库才会被创建，所有导出的类才可用


    Parameters:
    ----------
    path : str
        数据库路径，第一次连接时必须指定数据库路径
        后续使用如果不指定，会使用第一次指定的路径，如果再次指定，会覆盖第一次指定的路径
    autocreate : bool
        是否自动创建数据库，如果设置为True，当数据库不存在时，会自动创建数据库
        如果设置为False，当数据库不存在时，会抛出FileNotFoundError异常
        设置此参数是为了严格控制数据库创建行为，避免误操作
    wal : bool
        是否启用WAL模式，如果启用，可以提高并发性能，但会增加数据库文件大小
        WAL模式下，数据库文件会变大，但是读写性能会快很多

    Return:
    -------
    swandb :
        数据库实例

    Raises:
    -------
    FileNotFoundError :
        如果数据库不存在，并且autocreate为False，则会抛出FileNotFoundError异常
    ValueError :
        如果指定了路径但是没有设置autocreate为True，则会抛出ValueError异常
    """
    global db_path
    bound = True
    # 覆盖全局变量后需要重新创建数据库
    if not path and not db_path:
        raise ValueError("First time connect must specify the path")
    elif path:
        # 覆盖全局变量
        db_path = os.path.join(path, "runs.swanlab")
        bound = False
    # 检查数据库是否存在，不存在就创建
    db_exists = os.path.exists(db_path)
    if not db_exists and not autocreate:
        raise FileNotFoundError(f"DB file {db_path} not found")
    # 启用外键约束，启用WAL模式
    swandb = SqliteDatabase(db_path, pragmas={"foreign_keys": 1, "journal_mode": "wal"})
    if not bound:
        # 动态绑定数据库
        swandb.connect()
        swandb.bind(tables)
        swandb.create_tables(tables)
        swandb.close()
        if not Tag.field_exists("folder"):
            # 添加 folder 字段
            compat_tag_key(SqliteDatabase(db_path))
    return swandb



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/4/28 16:21
@File: pytest_file.py
@IDE: pycharm
@Description:
    测试文件格式函数
"""
import pytest
from nanoid import generate

# noinspection PyProtectedMember
from swanboard.utils import check_desc_format, check_exp_name_format
from swanboard.utils.file import _auto_cut


class TestAutoCut:
    @pytest.mark.parametrize(
        "name, value",
        [
            [generate(), generate(size=101)],
            [generate(), generate(size=1000)],
            [generate(), generate(size=10000)],
        ],
    )
    def test_cut(self, name: str, value: str):
        """
        测试自动截断
        """
        assert len(_auto_cut(name, value, 100, True)) == 100

    @pytest.mark.parametrize(
        "name, value",
        [
            [generate(), generate(size=101)],
            [generate(), generate(size=1000)],
            [generate(), generate(size=10000)],
        ],
    )
    def test_no_cut(self, name: str, value: str):
        """
        测试不自动截断
        """
        with pytest.raises(IndexError) as e:
            _auto_cut(name, value, 100, False)
        assert name in str(e.value)


class TestExpName:
    @pytest.mark.parametrize(
        "value",
        [
            generate(size=100),
            generate(size=1),
            "-",
            "_",
        ],
    )
    def test_exp_name_common(self, value):
        """
        测试正常情况
        """
        assert check_exp_name_format(value) == value

    @pytest.mark.parametrize("value", [None, 1, [], {}])
    def test_exp_name_type_error(self, value: str):
        """
        测试类型错误
        """
        with pytest.raises(TypeError):
            check_exp_name_format(value)

    @pytest.mark.parametrize("value", ["啊哈哈", "&^%", "/;]x]", generate(size=100), "👨‍💻👨‍💻👨‍💻👨‍💻👨‍"])
    def test_exp_name_value_special(self, value: str):
        """
        测试特殊值
        """
        assert check_exp_name_format(value) == value

    @pytest.mark.parametrize(
        "value",
        [
            "",
            "   ",
        ],
    )
    def test_exp_name_value_empty(self, value: str):
        """
        测试空值
        """
        with pytest.raises(ValueError):
            check_exp_name_format(value)

    @pytest.mark.parametrize(
        "value",
        [
            generate(size=101),
            generate(size=1000),
            generate(size=10000),
        ],
    )
    def test_exp_name_auto_cut(self, value: str):
        """
        测试自动截断
        """
        assert len(check_exp_name_format(value)) == 100

    @pytest.mark.parametrize(
        "value",
        [
            generate(size=101),
            generate(size=1000),
            generate(size=10000),
        ],
    )
    def test_exp_name_no_cut(self, value: str):
        """
        测试不自动截断
        """
        with pytest.raises(IndexError):
            check_exp_name_format(value, auto_cut=False)


class TestDesc:
    @pytest.mark.parametrize("value", [generate(size=100), generate(size=1), "-", "_", "👾👾👾👾👾👾"])
    def test_desc_common(self, value):
        """
        测试正常情况
        """
        assert check_desc_format(value) == value

    @pytest.mark.parametrize("value", [None, 1, [], {}])
    def test_desc_type_error(self, value: str):
        """
        测试类型错误
        """
        with pytest.raises(TypeError):
            check_desc_format(value)

    @pytest.mark.parametrize("value", ["", "   ", " " * 256])
    def test_desc_value_empty(self, value: str):
        """
        测试空值
        """
        assert check_desc_format(value) == ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-06-30 14:37:33
@File: test/unit/db/test_migrate.py
@IDE: vscode
@Description:
    测试兼容性迁移脚本
"""
from swanboard.db.migrate import compat_tag_key
from peewee import SqliteDatabase
from tutils import create_test_dir, mock_experiment
from swanboard.db import (
    Project,
    Tag,
)
from swankit.env import create_time
from urllib.parse import quote
import os
import nanoid


def init_db(db_path: str):
    # 模拟 项目=>实验=>tag
    project = Project.init("test_project_for_migrate")
    experiment_id = mock_experiment(project.id)
    db_path = os.path.join(db_path, "runs.swanlab")
    swandb = SqliteDatabase(db_path)
    # 删除 tag 表，重新创建一个没有 folder 字段的 tag 表
    swandb.execute_sql("DROP TABLE IF EXISTS tag")
    swandb.execute_sql(
        "CREATE TABLE tag (id INTEGER PRIMARY KEY AUTOINCREMENT, experiment_id INTEGER, name TEXT, type TEXT, description TEXT, system INTEGER, sort INTEGER, more TEXT, create_time TEXT, update_time TEXT)"
    )
    return swandb, experiment_id


def mock_tag_data(experiment_id: int, name: str):
    # 向 experiment 下添加 tag，该 tag 记录没有 folder 字段
    raw_sql = """
        INSERT INTO tag (experiment_id, name, type, description, system, sort, more, create_time, update_time)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
    """
    time = create_time()
    params = (
        experiment_id,  # experiment_id
        name,  # name
        "default",  # type
        "test_description",  # description
        0,  # system
        0,  # sort
        "",  # more
        time,  # create_time
        time,  # update_time
    )
    wrapper = Tag.raw(raw_sql, *params).execute()


class TestMigrateForTag:

    def test_compat_tag_key(self):
        db_path = create_test_dir("test_migrate")
        # 重新初始化一个数据库，其中 tag 表没有 folder 字段
        swandb, experiment_id = init_db(db_path)
        assert not Tag.field_exists("folder")
        # 模拟 tag 记录，并创建各自的独立存储目录
        list = [nanoid.generate(size=10), f"{nanoid.generate(size=4)}/{nanoid.generate(size=4)}"]
        with swandb.atomic():
            for name in list:
                mock_tag_data(experiment_id, name)
        compat_tag_key(swandb)
        # 含有 folder 字段，且值为转义后的 tag name
        assert Tag.field_exists("folder")
        for tag in Tag.select():
            assert tag.folder == quote(tag.name, safe="")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/4/3 16:52
@File: conftest.py.py
@IDE: pycharm
@Description:
    配置pytest
"""
import pytest
from tutils import clear, create_test_dir


@pytest.fixture(scope="session", autouse=True)
def setup_before_all():
    clear()


@pytest.fixture(scope="class", autouse=True)
def setup_before_class():
    create_test_dir()



from swankit.callback import SwanKitCallback, ColumnInfo
from .db.models import *
from .db import add_multi_chart, connect, NotExistedError, ExistedError, ChartTypeError
from typing import Tuple, Optional
from swanboard.utils import swanlog, get_swanlog_dir
import time
import re


class SwanBoardCallback(SwanKitCallback):
    """
    SwanBoardCallback类，swanlab本体与数据库的连接回调函数
    """

    def __init__(self):
        super(SwanBoardCallback, self).__init__()
        self.exp: Optional[Experiment] = None

    def __str__(self) -> str:
        return "SwanBoardCallback"

    def on_init(self, proj_name: str, *args, **kwargs):
        # 连接本地数据库，要求路径必须存在，但是如果数据库文件不存在，会自动创建
        connect(autocreate=True, path=get_swanlog_dir())
        # 初始化项目数据库
        Project.init(proj_name)

    def before_init_experiment(
        self,
        run_id: str,
        exp_name: str,
        description: str,
        num: int,
        colors: Tuple[str, str],
    ):
        pattern = r"-\d+$"
        # ---------------------------------- 实验名称校验 ----------------------------------
        # 这个循环的目的是如果创建失败则等零点五秒重新生成后缀重新创建，直到创建成功
        # 搜索所有以exp_name开头的实验
        count = Experiment.filter(Experiment.name.startswith(exp_name)).count()
        while True:
            try:
                # 获得数据库实例
                self.exp = Experiment.create(
                    name=exp_name,
                    run_id=run_id,
                    description=description,
                    num=num,
                    colors=colors,
                )
                break
            except ExistedError:
                count += 1
                swanlog.debug(f"Experiment {exp_name} has existed, try another name...")
                # 以-{数字}结尾
                if bool(re.search(pattern, exp_name)):
                    arr = exp_name.split("-")
                    arr[-1] = str(count)
                    exp_name = "-".join(arr)
                else:
                    exp_name = f"{exp_name}-{count}"
                time.sleep(0.2)

    def on_log(self):
        # 每一次log的时候检查一下数据库中的实验状态
        # 如果实验状态不为0，说明实验已经结束，不允许再次调用log方法
        # 这意味着每次log都会进行查询，比较消耗性能，后续考虑采用多进程共享内存的方式进行优化
        swanlog.debug(f"Check experiment and state...")
        try:
            exp = Experiment.get(id=self.exp.id)
        except NotExistedError:
            raise KeyboardInterrupt("The experiment has been deleted by the user")
        # 此时self.__state == 0，说明是前端主动停止的
        if exp.status != 0:
            raise KeyboardInterrupt("The experiment has been stopped by the user")

    def on_column_create(self, column_info: ColumnInfo):
        # 创建Chart
        chart = Chart.create(
            column_info.key,
            experiment_id=self.exp,
            type=column_info.chart.value.chart_type,
            reference=column_info.reference,
            config=column_info.config,
        )
        # 创建命名空间，如果命名空间已经存在，会抛出ExistedError异常，捕获不处理即可
        # 需要指定sort，default命名空间的sort为0，其他命名空间的sort为None，表示默认添加到最后
        namespace = column_info.namespace
        try:
            n = Namespace.create(name=namespace, experiment_id=self.exp.id, sort=column_info.sort)
            swanlog.debug(f"Namespace {namespace} created, id: {n.id}")
        except ExistedError:
            n: Namespace = Namespace.get(name=namespace, experiment_id=self.exp.id)
            swanlog.debug(f"Namespace {namespace} exists, id: {n.id}")
        # 创建display，这个必然是成功的，因为display是唯一的，直接添加到最后一条即可
        Display.create(chart_id=chart.id, namespace_id=n.id)
        tag: Tag = Tag.create(
            experiment_id=self.exp.id,
            name=column_info.key,
            type=column_info.chart.value.chart_type,
            folder=column_info.id,
        )
        # 添加一条source记录
        error = None
        if column_info.error is not None:
            error = {
                "data_class": column_info.error.got,
                "excepted": column_info.error.expected,
            }
        Source.create(tag_id=tag.id, chart_id=chart.id, error=error)
        # 新建多实验对比图表数据
        try:
            add_multi_chart(tag_id=tag.id, chart_id=chart.id)
        except ChartTypeError:
            swanlog.debug("In the multi-experiment chart, the current type of tag is not as expected.")

    def on_stop(self, error: str = None):
        # 更新数据库中的实验状态
        self.exp.update_status(-1 if error is not None else 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-16 10:52:10
@File: swanlab\db\__init__.py
@IDE: vscode
@Description:
    数据库模型模块，在设计上应该在data模块或者server模块被初始化完毕后导入
    导入之前必须确保文件路径存在，否则会报错
"""
from .models import (
    Project,
    Experiment,
    Tag,
    Chart,
    Namespace,
    Source,
    Display,
)
from .error import (
    ExistedError,
    NotExistedError,
    ForeignProNotExistedError,
    ForeignExpNotExistedError,
    ForeignTagNotExistedError,
    ForeignChartNotExistedError,
    ForeignNameNotExistedError,
    ChartTypeError,
)
from .table_config import tables
from .db_connect import connect
from .utils import add_multi_chart



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-07-01 20:41:20
@File: test/unit/setup/test_setup.py
@IDE: vscode
@Description:
    测试初始化函数
"""
from tutils import create_test_dir
import os


def test_create_test_dir():
    """测试创建测试文件夹"""
    path = create_test_dir("test_for_create_dir")
    assert os.path.exists(path)



from swanboard.utils.package import get_package_version
import pkg_resources


class TestGetPackageVersion:

    def test_get_package_version_installed(self, monkeypatch):
        """测试 swanlab 包已安装的情况"""

        def mock_get_distribution(name):
            class MockDistribution:
                version = "1.2.3"

            return MockDistribution()

        monkeypatch.setattr(pkg_resources, "get_distribution", mock_get_distribution)
        assert get_package_version() == "1.2.3"

    def test_get_package_version_not_installed(self, monkeypatch):
        """测试 swanlab 包未安装的情况"""

        def mock_get_distribution(name):
            raise pkg_resources.DistributionNotFound

        monkeypatch.setattr(pkg_resources, "get_distribution", mock_get_distribution)
        assert get_package_version() == "swanlab package is not installed"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-03-09 21:43:59
@File: swanlab/server/controller/utils/charts.py
@IDE: vscode
@Description:
    图表相关函数
"""
from ..db import Chart, Display, Namespace
from typing import List, Union


def get_exp_charts(id: int):
    """
    获取单实验图表数据

    Parameters
    ----------
    id : int
        实验id
    """
    charts: List[Chart] = Chart.filter(Chart.experiment_id == id)
    chart_list = Chart.search2list(charts)
    # 获取每个图表对应的数据源
    for index, chart in enumerate(charts):
        sources = []
        error = {}
        # source->experiment_id
        source_map = {}
        for source in Chart.search2list(chart.sources):
            sources.append(source["tag_id"]["name"])
            source_map[source["tag_id"]["name"]] = id
            if source["error"]:
                error[source["tag_id"]["name"]] = Chart.json2dict(source["error"])
        chart_list[index]["error"] = error
        chart_list[index]["source"] = sources
        chart_list[index]["multi"] = False
        chart_list[index]["source_map"] = source_map

    # 当前实验下的命名空间
    namespaces = Namespace.filter(Namespace.experiment_id == id)
    namespace_list = Namespace.search2list(namespaces)
    # 获取每个命名空间对应的 display
    # display 含有 chart 与 namespace 的对应关系
    for index, namespace in enumerate(namespaces):
        displays = []
        for display in Namespace.search2list(namespace.displays):
            displays.append(display["chart_id"]["id"])
        namespace_list[index]["charts"] = displays
    return get_pinned_and_hidden(chart_list, namespace_list)


def get_proj_charts(id: int):
    """
    获取多实验对比图表数据

    Parameters
    ----------
    id : int
        项目id
    """
    # 获取当前项目下所有的多实验对比表
    # 暂时只请求chart.type为default或者line的图表
    allow_types = ["default", "line", "image", "audio"]
    multi_charts = Chart.filter(Chart.project_id == id, Chart.type.in_(allow_types))
    # 获取图表配置
    charts = []
    # source -> experiment_id
    source_map = {}
    for _chart in multi_charts:
        # 多实验图表的 source 中不是 tag_name，而是 experiment_name
        sources, error = [], {}
        # 单箭头是通过外键反向索引的 chart -> source -> tag -> experiment => experiment_name
        for source in _chart.sources:
            sources.append(source.tag_id.experiment_id.name)
            source_map[source.tag_id.experiment_id.name] = source.tag_id.experiment_id.id
        # 当前chart的error字段，将所有的error字段转换为dict，key为实验名
        for source in _chart.sources:
            if source.error:
                error[source.tag_id.experiment_id.name] = Chart.json2dict(source.error)
        t = _chart.__dict__()
        charts.append({**t, "error": error, "source": sources, "multi": True, "source_map": source_map})
    # 获取命名空间配置
    namespaces = Namespace.filter(Namespace.project_id == id)
    namespace_list = Namespace.search2list(namespaces)
    # 通过 namespace 获取 display，从而获取该 namespace 下的 charts
    for index, namespace in enumerate(namespaces):
        displays = []
        for display in Display.search2list(namespace.displays):
            # 只获取默认图表
            if display["chart_id"]["type"] in allow_types:
                displays.append(display["chart_id"]["id"])
        if len(displays) > 0:
            namespace_list[index]["charts"] = displays
    # 过滤namespace中的空charts的namespace
    namespace_list = [namespace for namespace in namespace_list if "charts" in namespace]

    return get_pinned_and_hidden(charts, namespace_list)


def get_pinned_and_hidden(chart_list: List[dict], namespace_list: List[dict]) -> Union[List[dict], List[dict]]:
    """
    获取置顶图表与隐藏图表
    """
    if not len(chart_list) or not len(namespace_list):
        return chart_list, namespace_list
    # 获取pinned和hidden的开启/关闭状态，通过chart_list的第一个元素的experiment_id或者project_id获取
    first_chart = chart_list[0]
    if first_chart["experiment_id"] is not None:
        exp_or_proj = first_chart["experiment_id"]
    else:
        exp_or_proj = first_chart["project_id"]
    pinned_opened, hidden_opened = exp_or_proj["pinned_opened"], exp_or_proj["hidden_opened"]

    # 遍历chart_list，动态生成pinned与hidden的namespace，这两个namespace的id分别为-1与-2
    pinned_namespace = {
        "id": -1,
        "name": "pinned",
        "charts": [],
        "opened": pinned_opened,
        "experiment_id": first_chart["experiment_id"],
        "project_id": first_chart["project_id"],
    }
    hidden_namespace = {
        "id": -2,
        "name": "hidden",
        "charts": [],
        "opened": hidden_opened,
        "experiment_id": first_chart["experiment_id"],
        "project_id": first_chart["project_id"],
    }
    for chart in chart_list:
        # 如果chart的status为1，则将其加入pinned的namespace，如果是-1加入hidden的namespace
        # 如果是0，则不加入任何namespace
        # 首先将chart对象加入pinned或hidden的namespace，后续会滤除为id
        if chart["status"] == 1:
            pinned_namespace["charts"].append(chart)
        elif chart["status"] == -1:
            hidden_namespace["charts"].append(chart)
        if chart["status"] != 0:
            del_chart_from_namespace(namespace_list, chart["id"])
    # 滤除namespace中的空charts的namespace
    namespace_list = [namespace for namespace in namespace_list if len(namespace["charts"]) != 0]
    # 如果pinned_namespace中有charts，则加入namespace_list的首位
    if len(pinned_namespace["charts"]) > 0:
        # namespaces的charts字段根据每个元素的sort排序，小的在前
        pinned_namespace["charts"].sort(key=lambda x: x["sort"])
        pinned_namespace = {**pinned_namespace, "charts": [chart["id"] for chart in pinned_namespace["charts"]]}
        namespace_list.insert(0, pinned_namespace)
    # 如果hidden_namespace中有charts，则加入namespace_list的末位
    if len(hidden_namespace["charts"]) > 0:
        # namespaces的charts字段根据每个元素的sort排序，小的在前
        hidden_namespace["charts"].sort(key=lambda x: x["sort"])
        hidden_namespace = {**hidden_namespace, "charts": [chart["id"] for chart in hidden_namespace["charts"]]}
        namespace_list.append(hidden_namespace)
    return chart_list, namespace_list


def del_chart_from_namespace(namespace_list: List[dict], chart_id: int) -> List[dict]:
    """
    从命名空间中删除图表
    """
    for namespace in namespace_list:
        if chart_id in namespace["charts"]:
            namespace["charts"].remove(chart_id)
    return namespace_list



from swanboard.utils.font import *
import pytest
import time


class TestGenerateColor:

    def test_generate_color(self):
        # 测试边界情况
        assert generate_color(1) == (COLOR_LIST["light"][0], COLOR_LIST["dark"][0])
        assert generate_color(16) == (COLOR_LIST["light"][15], COLOR_LIST["dark"][15])
        # 测试循环情况
        assert generate_color(17) == (COLOR_LIST["light"][0], COLOR_LIST["dark"][0])
        assert generate_color(32) == (COLOR_LIST["light"][15], COLOR_LIST["dark"][15])
        # 测试其他任意数字
        assert generate_color(5) == (COLOR_LIST["light"][4], COLOR_LIST["dark"][4])
        assert generate_color(20) == (COLOR_LIST["light"][3], COLOR_LIST["dark"][3])


class TestFont:
    def test_loading(self):
        def mock_function(duration):
            time.sleep(duration)
            return "done"

        # 检查正常情况
        result = FONT.loading("Loading...", mock_function, (0.5,), interval=0.1)
        assert result == "done"

        # 检查异常情况
        with pytest.raises(Exception):
            FONT.loading("Loading...", mock_function, ("not a number",), interval=0.1)

    def test_swanlab(self):
        assert FONT.swanlab("Hello") == "\033[1m\033[34mswanlab\033[0m\033[0m: Hello"
        assert FONT.swanlab("Hello", "green") == "\033[1m\033[32mswanlab\033[0m\033[0m: Hello"

    def test_bold(self):
        assert FONT.bold("bold text") == "\033[1mbold text\033[0m"

    def test_default(self):
        assert FONT.default("default text") == "\033[39mdefault text\033[0m"

    def test_blue(self):
        assert FONT.blue("blue text") == "\033[34mblue text\033[0m"

    def test_grey(self):
        assert FONT.grey("grey text") == "\033[90mgrey text\033[0m"

    def test_underline(self):
        assert FONT.underline("underline text") == "\033[4munderline text\033[0m"

    def test_green(self):
        assert FONT.green("green text") == "\033[32mgreen text\033[0m"

    def test_dark_green(self):
        assert FONT.dark_green("dark green text") == "\033[38;5;22mdark green text\033[0m"

    def test_dark_gray(self):
        assert FONT.dark_gray("dark gray text") == "\033[38;5;236mdark gray text\033[0m"

    def test_yellow(self):
        assert FONT.yellow("yellow text") == "\033[33myellow text\033[0m"

    def test_red(self):
        assert FONT.red("red text") == "\033[31mred text\033[0m"

    def test_magenta(self):
        assert FONT.magenta("magenta text") == "\033[35mmagenta text\033[0m"

    def test_clear(self):
        colored_string = "\033[34mblue text\033[0m"
        assert FONT.clear(colored_string) == "blue text"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-29 14:14:02
@File: swanlab\server\router\static.py
@IDE: vscode
@Description:
    多媒体相关接口路由
    暂时只有一个，用于获取媒体数据，这媒体包括音频、视频、图片、文字等，返回的接口是统一的
    返回的数据格式也不再是JSON，而是二进制数据，这样可以减少数据传输的大小
    我们约定前端通过一个相对路径获取媒体数据，这个路径相对于当前watch的文件夹
"""

from ..settings import get_media_dir
from fastapi import APIRouter
from fastapi.responses import FileResponse
from ..db import Experiment, Tag
import os

router = APIRouter()


# ---------------------------------- 音频相关 ----------------------------------


@router.get("/{path:path}")
def _(path: str, tag: str, experiment_id: str):
    """获取媒体文件
    通过参数拼接为本地路径，返回二进制数据
    tag需要进行url编码
    """
    run_id = Experiment.get(Experiment.id == experiment_id).run_id
    tag_folder = Tag.filter(Tag.name == tag, Tag.experiment_id == experiment_id).first().folder
    media_path = os.path.join(get_media_dir(run_id, tag_folder), path)
    return FileResponse(media_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-16 10:54:30
@File: swanlab\db\modules\project.py
@IDE: vscode
@Description:
    项目表，对应于0.1.5之前版本中的 project.json
"""
from peewee import CharField, IntegerField, DatabaseProxy
from ..model import SwanModel
from swanboard.utils import get_package_version


class Project(SwanModel):
    """项目表
    目前，在一个工程中，只有一个项目
    """

    # 默认的项目id应该是1
    DEFAULT_PROJECT_ID = 1

    class Meta:
        database = DatabaseProxy()

    id = IntegerField(primary_key=True)
    """项目id"""
    name = CharField(max_length=100, null=False, unique=True)
    """项目名称，唯一"""
    description = CharField(max_length=255, null=True)
    """项目描述，可为空"""
    sum = IntegerField(null=True)
    """项目下实验数量，包括已删除的实验，这是一个只增不减的值"""
    charts = IntegerField(default=0, choices=[0, 1], null=False)
    """是否已经生成项目级别图表，0 未生成，1 已生成"""
    pinned_opened = IntegerField(default=1, choices=[0, 1])
    """多实验图表置顶部分是否打开，默认值为1，表示打开"""
    hidden_opened = IntegerField(default=0, choices=[0, 1])
    """多实验图表隐藏部分是否打开，默认值为0,表示关闭"""
    more = CharField(null=True)
    """更多信息配置，json格式"""
    version = CharField(max_length=30, null=False)
    """创建时的版本号"""
    create_time = CharField(max_length=30, null=False)
    """创建时间"""
    update_time = CharField(max_length=30, null=False)
    """更新时间"""

    def __dict__(self):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "sum": self.sum,
            "charts": self.charts,
            "more": self.more,
            "pinned_opened": self.pinned_opened,
            "hidden_opened": self.hidden_opened,
            "version": self.version,
            "create_time": self.create_time,
            "update_time": self.update_time,
        }

    @classmethod
    def init(cls, name="", description=None, more="") -> "Project":
        """
        静态方法
        初始化项目表，如果已经有项目存在，则不创建，直接返回第一条数据实例

        Parameters
        ----------
        name : str, optional
            项目名称，默认为空字符串
        description : str, optional
            项目描述，默认为空字符串
        more : str, optional
            更多信息配置，默认为空字符串

        Returns
        -------
        Project:
            项目实例
        """
        # 如果已经初始化，则不创建，直接返回第一条数据实例
        if cls.select().count() >= 1:
            return cls.filter(cls.id == cls.DEFAULT_PROJECT_ID)[0]
        # 创建项目
        return cls.create(
            name=name,
            description=description,
            more=more,
            version=get_package_version(),
        )

    @classmethod
    def increase_sum(cls, id=DEFAULT_PROJECT_ID) -> int:
        """
        静态方法
        更新实验统计数量，增加1
        此方法通常在创建实验时被Experiment类调用
        Parameters
        ----------
        id : int
            实验id, 默认为DEFAULT_PROJECT_ID

        Returns
        -------
        int:
            当前实验统计数量
        """
        project: "Project" = cls.filter(cls.id == id)[0]
        project.sum = project.sum + 1 if project.sum else 1
        project.save()
        return project.sum

    @classmethod
    def decrease_sum(cls, id=DEFAULT_PROJECT_ID) -> int:
        """
        静态方法
        更新实验统计数量，减少1
        此方法通常在删除实验时被Experiment类调用
        Parameters
        ----------
        id : int
            实验id, 默认为DEFAULT_PROJECT_ID

        Returns
        -------
        int:
            当前实验统计数量
        """
        project: "Project" = cls.filter(cls.id == id)[0]
        project.sum = project.sum - 1 if project.sum else 0
        project.save()
        return project.sum



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2023-12-13 19:46:18
@File: swanlab/server/module/resp.py
@IDE: vscode
@Description:
    定义响应结构体
    说是结构体，实际上是各种函数，用于返回各种响应
    结构体名称结构为：[错误描述]_[HTTP状态码]
"""
from fastapi.responses import JSONResponse as _JSONResponse

# ---------------------------------- 错误码 ----------------------------------


_SUCCESS_200 = 0
"""一切正常，成功，期望的HTTP状态码为200"""

_PARAMS_ERROR_422 = 3422
"""参数错误，通常对应着前端传输的参数无法通过校验，这在中间件中处理，期望的HTTP状态码为422"""

_NOT_FOUND_404 = 3404
"""资源不存在，通常对应着路径不存在，期望的HTTP状态码为404"""

_CONFLICT_409 = 3409
"""可能造成冲突，通常意味着试图对资源进行不合理的操作，期望的HTTP状态码为409"""

_DATA_ERROR_500 = 3500
"""服务端存储的数据格式错误，这通常意味着指定资源无法解析为期望格式，期望的HTTP状态码为500"""

_UNEXCEPTED_ERROR_500 = 3555
"""未知错误，通常对应着未知的异常，期望的HTTP状态码为500"""

# ---------------------------------- 定义响应结构体 ----------------------------------


def _ResponseBody(code: int, message: str = None, data: dict = None):
    """构造响应，返回一个字典，包含响应码，响应信息和响应数据

    Parameters
    ----------
    code : int
        响应码，0表示成功，非0表示失败
    message : str, optional
        错误信息，如果code为0，错误信息强制为success，如果code不为0，错误信息必须提供
    data : dict, optional
        响应数据，如果传入，必须为字典类型
    """
    # 如果code为0，错误信息强制为success
    message = "success" if code == 0 else message
    # 如果code不为0，错误信息必须提供
    assert code == 0 or message is not None and len(message) > 0
    # 如果传入了响应数据，必须为字典类型
    assert data is None or isinstance(data, dict)
    # 构造响应
    if data is None:
        return {
            "code": code,
            "message": message,
        }
    else:
        return {
            "code": code,
            "message": message,
            "data": data,
        }


def SUCCESS_200(data: dict):
    """成功响应"""
    return _JSONResponse(
        status_code=200,
        content=_ResponseBody(_SUCCESS_200, data=data),
    )


def PARAMS_ERROR_422(message: str = "params error"):
    """请求参数错误"""
    return _JSONResponse(
        status_code=422,
        content=_ResponseBody(_PARAMS_ERROR_422, message=message),
    )


def NOT_FOUND_404(message: str = "NotFound"):
    """资源不存在"""
    return _JSONResponse(
        status_code=404,
        content=_ResponseBody(_NOT_FOUND_404, message=message),
    )


def CONFLICT_409(message: str = "Conflict"):
    """请求会造成资源冲突"""
    return _JSONResponse(
        status_code=409,
        content=_ResponseBody(_CONFLICT_409, message=message),
    )


def DATA_ERROR_500(message: str = "data error"):
    """服务端存储的数据格式错误"""
    return _JSONResponse(
        status_code=500,
        content=_ResponseBody(_DATA_ERROR_500, message=message),
    )


def UNEXPECTED_ERROR_500(message: str = "unexpected error"):
    """未知错误"""
    return _JSONResponse(
        status_code=500,
        content=_ResponseBody(_UNEXCEPTED_ERROR_500, message=message),
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-18 13:52:35
@File: swanlab\db\models\chart.py
@IDE: vscode
@Description:
    实验图标表
"""
from ..model import SwanModel
from peewee import CharField, IntegerField, ForeignKeyField, TextField, IntegrityError, Check, DatabaseProxy, fn
from ..error import ExistedError, ForeignExpNotExistedError, ForeignProNotExistedError
from .experiments import Experiment
from .projects import Project


class Chart(SwanModel):
    """chart表，图表在namspace下的排序交由display表处理"""

    class Meta:
        database = DatabaseProxy()
        # 通过meta规定name和project_id的唯一性
        indexes = ((("name", "experiment_id"), True), (("name", "project_id"), True))

        constraints = [
            # project_id 和 experiment_id 不能同时为空，也不能同时不为空
            Check(
                "(project_id IS NULL AND experiment_id IS NOT NULL) OR (project_id IS NOT NULL AND experiment_id IS NULL)"
            ),
        ]

    id = IntegerField(primary_key=True)
    """图表id, 自增"""
    experiment_id = ForeignKeyField(Experiment, backref="charts", on_delete="CASCADE", on_update="CASCADE", null=True)
    """外键，关联的实验id，与project_id只有一个为NULL"""
    project_id = ForeignKeyField(
        Project, backref="charts", default=1, on_delete="CASCADE", on_update="CASCADE", null=True
    )
    """外键，关联的项目id，与experiment_id只有一个为NULL"""
    name = CharField(max_length=100, null=False)
    """图表名称"""
    description = CharField(max_length=255, null=True)
    """图表描述，可为空"""
    system = IntegerField(default=1, choices=[-1, 0, 1])
    """是否为创建tag时自动生成的图表，-1: 删除的自动生成的图表，0: 否，1: 是，系统图表不可删除，只能改为-1"""
    type = CharField(max_length=10, null=False)
    """图表类型，由创建者决定，这与数据库本身无关"""
    reference = CharField(max_length=10, default="step")
    """图表数据参考，由创建者决定，这与数据库本身无关"""
    status = IntegerField(null=False, default=0)
    """图表状态 0代表默认状态 1代表被pinned -1代表被hidden"""
    sort = IntegerField(null=True, default=None)
    """图表在被pinned或hidden时的排序，越小越靠前，越大越靠后，如果为NULL则代表未被pinned或hidden"""
    config = TextField(null=True)
    """图表的其他配置，这实际上是一个json字符串"""
    more = TextField(null=True)
    """更多信息配置，json格式"""
    create_time = CharField(max_length=30, null=False)
    """创建时间"""
    update_time = CharField(max_length=30, null=False)
    """更新时间"""

    def __dict__(self):
        experiment_id = None if self.experiment_id is None else self.experiment_id.__dict__()
        project_id = None if self.project_id is None else self.project_id.__dict__()
        return {
            "id": self.id,
            "experiment_id": experiment_id,
            "project_id": project_id,
            "name": self.name,
            "description": self.description,
            "system": self.system,
            "type": self.type,
            "reference": self.reference,
            "config": self.config,
            "status": self.status,
            "sort": self.sort,
            "more": self.more,
            "create_time": self.create_time,
            "update_time": self.update_time,
        }

    @classmethod
    def create(
        cls,
        name: str,
        description: str = None,
        experiment_id: int = None,
        project_id: int = None,
        system: int = 1,
        type: str = "default",
        reference: str = "step",
        config: dict = None,
        more: dict = None,
    ) -> "Chart":
        """创建实验图表

        Parameters
        ----------
        name : str
            实验图表名称
        description : str
            实验图表描述
        experiment_id : int, optional
            外键，关联的实验id，与project_id只有一个为NULL, 默认为None，但与project_id只有一个为None
        project_id : int, optional
            外键，关联的项目id，与experiment_id只有一个为NULL, 默认为None，但与experiment_id只有一个为None
        system : int, optional
            是否为创建tag时自动生成的图表，-1: 删除的自动生成的图表，0: 否，1: 是，系统图表不可删除，只能改为-1, 默认为是
        type : str, optional
            图表类型，由创建者决定，这与数据库本身无关, 默认为"default"
        reference : str, optional
            图表数据参考，由创建者决定，这与数据库本身无关, 默认为"step"
        config : dict, optional
            图表的其他配置，这实际上是一个json字符串, 默认为None
        more : dict, optional
            更多信息配置，json格式, 默认为None

        Returns
        -------
        Chart : Chart
            创建的实验图表

        Raises
        -------
        ForeignExpNotExistedError
            实验不存在
        ForeignProNotExistedError
            项目不存在
        ExistedError
            图标必须唯一
        """

        # 检查外键存在性
        if project_id and not Project.filter(Project.id == project_id).exists():
            raise ForeignProNotExistedError("项目不存在")
        if experiment_id and not Experiment.filter(Experiment.id == experiment_id).exists():
            raise ForeignExpNotExistedError("实验不存在")

        try:
            return super().create(
                experiment_id=experiment_id,
                project_id=project_id,
                name=name,
                description=description,
                system=system,
                type=type,
                reference=reference,
                config=config,
                more=more,
            )
        except IntegrityError:
            raise ExistedError("图表已存在")

    @classmethod
    def pin(cls, id: int, sort: int = None) -> "Chart":
        """
        把某个chart置顶

        Parameters
        ----------
        id : int
            图表id
        sort : int
            图表在被pinned或hidden时的排序，越小越靠前，越大越靠后，如果为NULL则代表未被pinned或hidden, 默认为None
            默认排在最后，也就是传入None的时候

        Returns
        -------
        Chart : Chart
            置顶的图表
        """
        chart: Chart = cls.get(cls.id == id)
        chart.status = 1
        if sort is None:
            # 判断当前实验/项目下有多少个hidden的chart
            project_id = None if chart.project_id is None else chart.project_id.id
            experiment_id = None if chart.experiment_id is None else chart.experiment_id.id
            # 获取当前namespace下的最大索引,如果没有，则为0
            sort = (
                cls.select(fn.Max(cls.sort))
                .where(cls.project_id == project_id, cls.experiment_id == experiment_id, cls.status == 1)
                .scalar()
            )
            sort = 0 if sort is None else sort + 1
        chart.sort = sort
        chart.save()
        return chart

    @classmethod
    def restore(cls, id: int) -> "Chart":
        """
        把某个chart恢复正常状态

        Parameters
        ----------
        id : int
            图表id

        Returns
        -------
        Chart : Chart
            恢复的图表
        """
        chart: Chart = cls.get(cls.id == id)
        chart.status = 0
        chart.sort = None
        chart.save()
        return chart

    @classmethod
    def hide(cls, id: int, sort: int = None) -> "Chart":
        """
        把某个chart隐藏

        Parameters
        ----------
        id : int
            图表id
        sort : int
            图表在被pinned或hidden时的排序，越小越靠前，越大越靠后，如果为NULL则代表未被pinned或hidden, 默认为None
            默认排在最后，也就是传入None的时候

        Returns
        -------
        Chart : Chart
            隐藏的图表
        """
        chart: Chart = cls.get(cls.id == id)
        chart.status = -1
        if sort is None:
            # 判断当前实验/项目下有多少个hidden的chart
            project_id = None if chart.project_id is None else chart.project_id.id
            experiment_id = None if chart.experiment_id is None else chart.experiment_id.id
            # 获取当前namespace下的最大索引,如果没有，则为0
            sort = (
                cls.select(fn.Max(cls.sort))
                .where(cls.project_id == project_id, cls.experiment_id == experiment_id, cls.status == -1)
                .scalar()
            )
            sort = 0 if sort is None else sort + 1
        chart.sort = sort
        chart.save()
        return chart



import os
import uvicorn
from swanboard.app import app
from swanboard.db import connect
from swanboard.utils import FONT

path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "swanlog")
connect(path=path)
HOST = "0.0.0.0"
PORT = 6092


def main():
    return app


if __name__ == "__main__":
    print("Try to explore the swanlab experiment logs in: \n")
    print("               --> " + FONT.dark_green("http://127.0.0.1:" + str(PORT)))
    print("\n")
    uvicorn.run("start_server:main", host=HOST, port=PORT, reload=True, log_level="critical")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-06-10 19:43:56
@File: swanboard/run.py
@IDE: vscode
@Description:
    启动服务
"""
import uvicorn
from .utils import URL, is_ipv4, is_port
import time
from swanboard.utils import FONT, swanlog as swl, get_package_version
from swanboard.db import connect
import sys


class SwanBoardRun:
    """swanboard 面板启动类"""

    @staticmethod
    def is_valid_ip(ip: str) -> None:
        """检测输入的是否是合法的ip地址,完成环境变量的注入

        Parameters
        ----------
        ctx : click.Context
            上下文
        param : click.Parameter
            参数
        ip : str
            带检测的字符串
        """
        if ip is None:
            return
        if not is_ipv4(ip):
            raise ValueError("Invalid ip address: " + ip)

    @staticmethod
    def is_valid_port(port: int) -> int:
        """检测是否是合法的端口号

        Parameters
        ----------
        ctx : click.Context
            上下文
        param : click.Parameter
            参数
        port : int
            带检测的端口号
        """
        if port is None:
            return
        if not is_port(port):
            raise ValueError("Invalid port number: " + str(port))

    @staticmethod
    def run(path: str, host: str, port: int) -> None:
        """启动函数

        Parameters
        ----------
        path : str
            swanlog 绝对路径
        host : str
            启动地址
        port : int
            启动端口
        """

        start = time.time()
        # debug一下当前日志文件夹的位置
        swl.debug("Try to explore the swanlab experiment logs in: " + FONT.bold(path))
        try:
            connect(path=path)
        except FileNotFoundError:
            swl.error("Can not find the swanlab db in: " + FONT.bold(path))
        # ---------------------------------- 日志打印 ----------------------------------
        # 可用URL
        ipv4 = URL.get_all_ip()
        if URL.is_zero_ip(host):
            tip = "\n".join([URL(i, port).__str__() for i in ipv4])
        else:
            tip = URL(host, port).__str__()
        tip = tip + "\n" + URL.last_tip() + "\n"
        v = FONT.bold("v" + get_package_version())
        # 耗时
        take_time = int((time.time() - start) * 1000).__str__() + "ms\n\n"
        swl.info(f"SwanLab Experiment Dashboard " + v + " ready in " + FONT.bold(take_time) + tip)
        # ---------------------------------- 启动服务 ----------------------------------
        # 使用 uvicorn 启动 FastAPI 应用，关闭原生日志
        # 使用try except 捕获退出，涉及端口占用等
        from swanboard.app import app

        try:
            uvicorn.run(app, host=host, port=port, log_level="critical")
        except SystemExit as e:
            code = e.code
            if code == 1:
                critical = "Error while attempting to bind on address ({}, {}): address already in use".format(
                    host, port
                )
                swl.critical(critical)
                sys.exit(7)
            else:
                swl.critical("Unhandled Exit Code: {}".format(code))
                sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-06-15 15:50:00
@File: swanboard/utils/package.py
@IDE: vscode
@Description:
    包相关的工具函数
"""
from importlib.metadata import version


def get_package_version() -> str:
    """获取swanlab的版本号

    Parameters
    ----------
    p : str, optional
        package.json文件路径，默认为项目的package.json

    Returns
    -------
    str
        swanlab的版本号
    """
    # 获取已安装包的版本
    try:
        v = version('swanlab')
        return v if v else 'unknown'
    except Exception as e:  # noqa
        return 'unknown'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-06-26 15:16:04
@File: test/unit/db/pytest_project.py
@IDE: vscode
@Description:
    测试项目数据表模型
"""
from tutils import mock_project
from swanboard.db import Project
import nanoid


class TestProject:

    def test_init(self):
        """测试初始化一个默认项目，如果项目已存在则跳过"""
        name = nanoid.generate(size=10)
        p = Project.init(name)
        project = Project.filter(Project.id == p.id).first()
        assert project.name == name
        assert project.sum == None
        assert project.charts == 0
        assert project.pinned_opened == 1
        assert project.hidden_opened == 0
        p2 = Project.init(name)
        assert p2.id == p.id

    def test_increase_sum(self):
        """测试增加项目实验数"""
        project_id = mock_project()
        project = Project.filter(Project.id == project_id).first()
        assert project.sum == 0
        assert project.increase_sum(project_id) == 1
        assert project.increase_sum(project_id) == 2
        project = Project.filter(Project.id == project_id).first()
        assert project.sum == 2

    def test_decrease_sum_with_zero(self):
        """测试减少项目实验数，不允许出现负数"""
        project_id = mock_project()
        project = Project.filter(Project.id == project_id).first()
        assert project.sum == 0
        assert project.decrease_sum(project_id) == 0
        project = Project.filter(Project.id == project_id).first()
        assert project.sum == 0

    def test_decrease_sum_with_many(self):
        """正常情况下，测试减少项目实验数"""
        project_id = mock_project()
        project = Project.filter(Project.id == project_id).first()
        assert project.sum == 0
        assert project.increase_sum(project_id) == 1
        assert project.increase_sum(project_id) == 2
        assert project.decrease_sum(project_id) == 1
        project = Project.filter(Project.id == project_id).first()
        assert project.sum == 1

    def test_2dict(self):
        """将实例转成字典"""
        project_id = mock_project()
        project = Project.filter(Project.id == project_id).first()
        assert isinstance(project.__dict__(), dict)



from ..module.resp import SUCCESS_200, NOT_FOUND_404, PARAMS_ERROR_422
from .db import NotExistedError

from .db import (
    Namespace,
    Experiment,
    Project,
)


# 修改namespace可见性
def change_namespace_opened(namespace_id: int, opened: int, experiment_id: int, project_id: int):
    """修改namespace的打开/关闭状态

    Parameters
    ----------
    namespace_id : int
        namespace的id，特别的，-1表示pinned，-2表示hidden
    opened : int
        在一个实验下，该namespace是否可见，true -> 1 , false -> 0
    experiment_id : int
        实验id，修改pinned/hidden的namespace的开启关闭状态时需要
    project_id : int
        项目id，修改pinned/hidden的namespace的开启关闭状态时需要
    Returns
    -------
    _type_:dict
        当前namespace信息
    """
    if namespace_id in [-1, -2]:
        if experiment_id is not None:
            exp_or_proj = Experiment.get_by_id(experiment_id)
        else:
            exp_or_proj = Project.get_by_id(project_id)
        # 如果实验或者项目不存在，则返回404
        if exp_or_proj is None:
            return NOT_FOUND_404("Experiment or project with id {} does not exist.".format(experiment_id))
        if namespace_id == -1:
            exp_or_proj.pinned_opened = 1 if opened else 0
        else:
            exp_or_proj.hidden_opened = 1 if opened else 0
        exp_or_proj.save()
        return SUCCESS_200(None)
    # 正常的namespace
    try:
        namespace: Namespace = Namespace.get_by_id(namespace_id)
    except NotExistedError:
        return NOT_FOUND_404("Namespace with id {} does not exist.".format(namespace_id))
    namespace.opened = 1 if opened else 0
    namespace.save()
    return SUCCESS_200(None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-06-26 15:17:56
@File: test/unit/db/pytest_experiment.py
@IDE: vscode
@Description:
    测试实验数据表模型
"""
import pytest
import nanoid
import os
from tutils import mock_project, mock_experiment, mock_folder_with_run_id
from swanboard.db import Experiment
from swanboard.db.error import ExistedError, NotExistedError


class TestExperiment:

    @pytest.fixture(scope="class", autouse=True)
    def setup(self):
        """在测试实验之前，需要一个默认项目"""
        project_id = mock_project()
        assert project_id == 1

    def test_create(self):
        """正常创建一个实验"""
        name = nanoid.generate(size=10)
        run_id = nanoid.generate(size=10)
        e = Experiment.create(name=name, run_id=run_id)
        assert e.name == name
        assert e.run_id == run_id
        assert e.project_id.id == 1

    def test_duplicate_name(self):
        """实验名重复"""
        name = nanoid.generate(size=10)
        run_id = nanoid.generate(size=10)
        e = Experiment.create(name=name, run_id=run_id)
        with pytest.raises(ExistedError):
            Experiment.create(name=name, run_id=nanoid.generate(size=10))

    def test_duplicate_run_id(self):
        """实验run_id重复"""
        name = nanoid.generate(size=10)
        run_id = nanoid.generate(size=10)
        e = Experiment.create(name=name, run_id=run_id)
        with pytest.raises(ExistedError):
            Experiment.create(name=nanoid.generate(size=10), run_id=run_id)

    def test_get_experiment(self):
        """正常获取一个实验"""
        e_id = mock_experiment()
        assert Experiment.get(e_id).id == e_id

    def test_get_nonexistent_experiment(self):
        """获取一个不存在的实验"""
        with pytest.raises(NotExistedError):
            Experiment.get(10000)

    def test_update_status_to_finished(self):
        """更新实验状态为结束"""
        e = Experiment.create(name=nanoid.generate(size=10), run_id=nanoid.generate(size=10))
        assert e.status == 0
        e.update_status(1)
        target = Experiment.get(e.id)
        assert target.status == 1

    def test_update_status_to_crashed(self):
        """更新实验状态为结束"""
        e = Experiment.create(name=nanoid.generate(size=10), run_id=nanoid.generate(size=10))
        assert e.status == 0
        e.update_status(-1)
        target = Experiment.get(e.id)
        assert target.status == -1

    def test_purely_delete(self):
        """正常删除一个实验"""
        name = nanoid.generate(size=10)
        run_id = nanoid.generate(size=10)
        e = Experiment.create(name=name, run_id=run_id)
        # 创建实验对应目录
        path = mock_folder_with_run_id(run_id)
        assert os.path.exists(path) is True
        e.purely_delete(run_id)
        # 删除之后 path 将不再存在
        assert os.path.exists(path) is False



from fastapi import APIRouter
from ..module.resp import PARAMS_ERROR_422

from ..controller.chart import (
    update_charts_status,
)
from fastapi import Request
from urllib.parse import quote

router = APIRouter()


@router.patch("/{chart_id}/status")
async def _(chart_id: int, request: Request):
    """修改图表状态

    Parameters
    ----------
    chart_id : int
        图表id
    request : Request
        请求体，包含 status 字段
    Returns
    -------
    groups : list
        图表列表
    """

    data = await request.json()
    status = data.get("status")

    if status is None:
        return PARAMS_ERROR_422("Request parameter 'status'")
    return update_charts_status(chart_id, status)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-17 20:43:22
@File: swanlab\db\modules\tag.py
@IDE: vscode
@Description:
    实验 tag 表
"""
from ..model import SwanModel
from peewee import ForeignKeyField, CharField, TextField, IntegerField, IntegrityError, DatabaseProxy, fn
from ..error import ExistedError, ForeignExpNotExistedError
from .experiments import Experiment


class Tag(SwanModel):
    """实验tag数据表

    Parameters
    ----------
    SwanModel : class
        自封装类
    """

    class Meta:
        database = DatabaseProxy()
        # 写入check约束，name和experiment_id加起来唯一
        indexes = ((("name", "experiment_id"), True),)

    id = IntegerField(primary_key=True)
    """tag唯一id"""
    # backref是反向引用，可以通过实验获取tag，比如experiment.tags获取此实验下的所有tag
    experiment_id = ForeignKeyField(Experiment, backref="tags", null=False, on_delete="CASCADE", on_update="CASCADE")
    """tag对应的实验id，代表这个tag由谁创建"""
    name = CharField(max_length=255, null=False)
    """tag名称，同一个实验下，tag名称不能重复"""
    folder = CharField(max_length=255, null=True)
    """tag目录名"""
    type = CharField(max_length=10, null=False)
    """tag的类型"""
    description = CharField(max_length=100, null=True)
    """tag的描述，可为空"""
    system = IntegerField(default=0, choices=[0, 1])
    """标识这个tag数据由系统生成还是用户生成，0: 用户生成，1: 系统生成，默认为0"""
    sort = IntegerField(default=0)
    """tag在实验中的排序，值越小越靠前"""
    more = TextField(null=True)
    """更多信息配置，json格式，将在表函数中检查并解析"""
    create_time = CharField(max_length=30, null=False)
    """tag的创建时间"""
    update_time = CharField(max_length=30, null=False)
    """tag的更新时间"""

    def __dict__(self):
        return {
            "id": self.id,
            "experiment_id": self.experiment_id,
            "name": self.name,
            "description": self.description,
            "system": self.system,
            "more": self.more,
            "create_time": self.create_time,
            "update_time": self.update_time,
        }

    @classmethod
    def create(
        cls,
        experiment_id: int,
        name: str,
        folder: str,
        type: str,
        description="",
        system: int = 0,
        more: dict = None,
    ) -> "Tag":
        """在tags表中创建一条tag指定实验的tag数据

        Parameters
        ----------
        experiment_id : int
            这将成为tag数据的外键，指向实验表中的某一条实验数据
        name : str
            tag的名称，字符串类型，小于255个字符
        type : str
            tag的类型，字符串类型，小于10个字符
        description : str, optional
            tag的描述，字符串类型，小于100个字符, 默认为空字符串
        system : int, optional
            判断是否是系统自动生成的tag，默认为否
        more : str, optional
            这必须是字典格式数据，代表更多信息, 默认为空字典

        Returns
        -------
        Tag
            返回创建的tag数据实例，可以通过实例获取tag数据

        Raises
        ------
        ForeignExpNotExistedError
            实验对应的id不存在
        ExistedError
            此tag已经在数据库中存在（tag-experiment_id唯一）
        """
        # 如果实验id不存在，则抛出异常
        if not Experiment.filter(Experiment.id == experiment_id).exists():
            raise ForeignExpNotExistedError("experiment不存在")
        # 获取当前实验下tag的最大排序索引，如果没有则为0
        sort = Tag.select(fn.Max(Tag.sort)).where(Tag.experiment_id == experiment_id).scalar()
        sort = sort + 1 if sort is not None else 0
        # 尝试创建实验tag，如果已经存在则抛出异常
        try:
            return super().create(
                experiment_id=experiment_id,
                name=name,
                folder=folder,
                type=type,
                description=description,
                system=system,
                sort=sort,
                more=more,
            )
        except IntegrityError:
            raise ExistedError("tag已存在")



from fastapi import APIRouter, Request

from ..controller.project import (
    # 列出实验列表
    get_project_info,
    # 获取项目总结信息
    get_project_summary,
    # 修改项目信息
    update_project_info,
    # 删除项目
    delete_project,
    # 获取多实验对比图表
    get_project_charts,
    # 默认项目id
    DEFAULT_PROJECT_ID,
)

router = APIRouter()


@router.get("")
def _():
    """获取项目下的实验列表

    Returns
    -------
    list[dict]:
        项目下的实验列表
    """

    return get_project_info()


@router.get("/summaries")
def _():
    """获取项目总结信息

    Returns
    -------
    dict
        项目下每个实验的总结信息
    """

    return get_project_summary()


@router.patch("/update")
async def _(request: Request, project_id: int = None):
    """修改项目信息

    Parameters
    ----------
    request : Request
        请求体，应该包含 name 和 description 字段
    project_id : int, optional
        by default None

    Returns
    -------
    dict
        修改后的项目信息
        - name
        - description
    """

    if project_id is None:
        return await update_project_info(request)
    return await update_project_info(request, project_id)


@router.delete("")
async def _(project_id: int = None):
    """删除项目

    Parameters
    ----------
    project_id : int, optional
        默认为第一个项目
    """

    if project_id is None:
        return await delete_project()
    return await delete_project(project_id)


@router.get("/charts")
async def _(project_id: int = DEFAULT_PROJECT_ID):
    """获取多实验对比图表数据,并且考虑往期版本兼容性
    1. 如果当前项目的chart字段为0，先生成多实验对比数据，跳转步骤2
    2. 依据规则获取所有实验的图表数据
    """

    return await get_project_charts(project_id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-18 15:16:55
@File: swanlab/db/error.py
@IDE: vscode
@Description:
    在此处定义所有的数据错误类型，以便于在其他地方引用
"""


class ExistedError(Exception):
    """某条数据已存在

    Parameters
    ----------
    Exception : class
        python内置异常类
    """

    pass


class NotExistedError(Exception):
    """某条数据不存在

    Parameters
    ----------
    Exception : class
        python内置异常类
    """

    pass


class ForeignExpNotExistedError(NotExistedError):
    """实验外键不存在"""

    pass


class ForeignProNotExistedError(NotExistedError):
    """项目外键不存在"""

    pass


class ForeignNameNotExistedError(NotExistedError):
    """命名空间外键不存在"""

    pass


class ForeignChartNotExistedError(NotExistedError):
    """图表外键不存在"""

    pass


class ForeignTagNotExistedError(NotExistedError):
    """标签外键不存在"""

    pass


class ChartTypeError(Exception):
    """图表类型错误，对应的"""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-20 21:21:45
@File: swanlab\server\controller\experiment.py
@IDE: vscode
@Description:
    实验相关 api 的处理函数
"""

import os
import ujson
import shutil
from datetime import datetime
from ..module.resp import SUCCESS_200, DATA_ERROR_500, NOT_FOUND_404
from fastapi import Request
from urllib.parse import quote
from ..settings import (
    get_logs_dir,
    get_exp_dir,
    get_config_path,
    get_console_dir,
    get_meta_path,
    get_requirements_path,
)
import yaml
from swanboard.utils import swanlog
from swankit.env import create_time
from .db import (
    Project,
    Experiment,
    Chart,
    Tag,
    connect,
    NotExistedError,
)
from .utils import (
    get_exp_charts,
    clear_field,
    read_tag_data,
    get_tag_files,
    LOGS_CONFIGS,
    lttb,
)

__to_list = Experiment.search2list

# ---------------------------------- 通用 ----------------------------------


# 默认项目 id
DEFAULT_PROJECT_ID = Project.DEFAULT_PROJECT_ID
# 实验运行状态
RUNNING_STATUS = Experiment.RUNNING_STATUS


# ---------------------------------- 工具函数 ----------------------------------


def __get_exp_dir_by_id(experiment_id: int) -> str:
    """通过 experiment_id 获取实验目录

    Parameters
    ----------
    experiment_id : int
        实验唯一id

    Returns
    -------
    str
        实验目录路径
    """

    return get_exp_dir(Experiment.get(experiment_id).run_id)


def __get_logs_dir_by_id(experiment_id: int) -> str:
    """通过 experiment_id 获取实验 保存目录

    Parameters
    ----------
    experiment_id : int
        实验唯一id

    Returns
    -------
    str
        实验 logs 目录路径
    """
    return get_logs_dir(Experiment.get(experiment_id).run_id)


def __get_console_dir_by_id(experiment_id: int) -> str:
    """通过 experiment_id 获取实验 console 保存目录

    Parameters
    ----------
    experiment_id : int
        实验唯一id

    Returns
    -------
    str
        实验 console 目录路径
    """

    return get_console_dir(Experiment.get(experiment_id).run_id)


def __get_requirements_path_by_id(experiment_id: int):
    """通过 experiment_id 获取实验依赖存储路径

    Parameters
    ----------
    experiment_id : int
        实验唯一id

    Returns
    -------
    str
        实验依赖存储路径
    """

    return get_requirements_path(Experiment.get(experiment_id).run_id)


# ---------------------------------- 路由对应的处理函数 ----------------------------------


# 获取实验信息
def get_experiment_info(experiment_id: int):
    """获取实验信息
    1. 数据库中获取实验的基本信息
    2. 从实验目录获取配置信息
    3. 获取实验元信息

    Parameters
    ----------
    experiment_id : int
        实验唯一id
    """

    experiment = Experiment.get(experiment_id).__dict__()
    experiment.pop("project_id")

    # 加载实验配置
    config_path = get_config_path(experiment["run_id"])
    if os.path.exists(config_path):
        with open(config_path, "r+", encoding="utf-8") as f:
            experiment["config"] = yaml.load(f, Loader=yaml.FullLoader)

    # 加载实验元信息
    meta_path = get_meta_path(experiment["run_id"])
    if os.path.exists(meta_path) and not os.stat(meta_path).st_size == 0:
        with open(meta_path, "r+", encoding="utf-8") as f:
            experiment["system"] = ujson.load(f)
    else:
        experiment["system"] = {}
    return SUCCESS_200(experiment)


# 获取表单数据
def get_tag_data(experiment_id: int, tag: str) -> dict:
    """获取表单数据
    根据实验id得到实验的运行id，然后根据运行id和tag得到实验的数据

    Parameters
    ----------
    experiment_id: int
        实验唯一id，路径传参
    tag: str
        表单标签，路径传参，已经进行了 URIComponent 编码
    """
    # ---------------------------------- 前置处理 ----------------------------------
    tag_folder = Tag.filter(Tag.name == tag, Tag.experiment_id == experiment_id).first().folder
    # 获取tag对应的存储目录
    try:
        tag_path: str = os.path.join(__get_logs_dir_by_id(experiment_id), tag_folder)
    except NotExistedError:
        return NOT_FOUND_404("experiment not found")
    if not os.path.exists(tag_path):
        return NOT_FOUND_404("tag not found")
    # 获取目录下存储的所有数据
    tag_data: list = []
    # ---------------------------------- 读取文件数据 ----------------------------------
    current_logs = get_tag_files(tag_path, LOGS_CONFIGS)
    for file in current_logs:
        tag_data = tag_data + read_tag_data(os.path.join(tag_path, file))
    # ---------------------------------- 返回所有数据 ----------------------------------
    # 如果数据为空，返回空列表
    if len(tag_data) == 0:
        return SUCCESS_200(data={"sum": 0, "list": [], "experiment_id": experiment_id})
    # 根据index升序排序
    tag_data.sort(key=lambda x: int(x["index"]))
    # tag_data 的 最后一个数据增加一个字段_last = True
    tag_data[-1]["_last"] = True
    # 获取_summary文件
    summary_path = os.path.join(tag_path, "_summary.json")
    if os.path.exists(summary_path):
        with open(summary_path, "r", encoding="utf-8") as f:
            summary = ujson.load(f)
            max_value = summary.get("max", None)
            min_value = summary.get("min", None)
    else:
        # COMPAT 如果_summary文件不存在，手动获取最大值和最小值
        warn = f"Summary file of tag '{tag}' not found, SwanLab will automatically get the maximum and minimum values."
        swanlog.warning(warn)
        # 遍历tag_data，获取最大值和最小值
        # 提取 data 字段的值
        data_values = [entry["data"] for entry in tag_data]
        # 获取最大值和最小值
        max_value = max(data_values)
        min_value = min(data_values)
    return SUCCESS_200(
        data={
            "sum": len(tag_data),
            "max": max_value,
            "min": min_value,
            "list": lttb(tag_data),
            # 标注此数据隶属于哪个实验
            "experiment_id": experiment_id,
        }
    )


# 获取实验状态
def get_experiment_status(experiment_id: int):
    """获取实验状态以及实验图表配置，用于实时更新实验状态

    Parameters
    ----------
    experiment_id : int
        实验唯一id，路径传参
    """

    experiment = Experiment.get(experiment_id)
    chart_list, namespace_list = get_exp_charts(experiment_id)

    return SUCCESS_200(
        {
            "status": experiment.status,
            "update_time": experiment.update_time,
            "finish_time": experiment.finish_time,
            "charts": {
                "charts": clear_field(chart_list, "experiment_id"),
                "namespaces": namespace_list,
            },
        }
    )


# 获取实验的总结数据
def get_experiment_summary(experiment_id: int) -> dict:
    """获取实验的总结数据——每个tag的最后一个setp的data

    Parameters
    ----------
    experiment_id : int
        实验id

    Returns
    -------
    dict:
        summaries: list
            每个tag的最后一个数据
    """

    experiment = Experiment.get_by_id(experiment_id)
    # 通过外键反链获取实验下的所有tag
    tag_list = [(tag["folder"], tag["name"]) for tag in __to_list(experiment.tags)]
    experiment_path = __get_logs_dir_by_id(experiment_id)
    # 通过目录结构获取所有正常的tag
    tags = [f for f in os.listdir(experiment_path) if os.path.isdir(os.path.join(experiment_path, f))]
    # 实验总结数据
    summaries = []
    for tag_key, tag_name in tag_list:
        # 如果 tag 记录在数据库，但是没有对应目录，说明 tag 有问题
        # 所以 tags 是 tag_list 的子集，出现异常的 tag 会记录在数据库但不会添加到目录结构中
        if tag_key not in tags:
            summaries.append({"key": tag_name, "value": "TypeError"})
            continue
        tag_path = os.path.join(experiment_path, tag_key)
        # 获取 tag 目录下的所有存储的日志文件
        logs = get_tag_files(tag_path, LOGS_CONFIGS)
        # 打开 tag 目录下最后一个存储文件，获取最后一条数据
        with open(os.path.join(tag_path, logs[-1]), encoding="utf-8") as f:
            lines = f.readlines()
            # 最后一行数据，如果为空，取倒数第二行
            data = lines[-1] if len(lines[-1]) else lines[-2]
            last_data = ujson.loads(data)
            summaries.append({"key": tag_name, "value": last_data["data"]})
    # 获取数据库记录时在实验下的排序
    sorts = {item["name"]: item["sort"] for item in __to_list(experiment.tags)}
    # 新版添加排序后的 tag，这里进行排序
    temp = [0] * len(summaries)
    for item in summaries:
        temp[sorts[item["key"]]] = item
    summaries = temp
    return SUCCESS_200({"summaries": summaries})


MAX_NUM = 6000


# 获取实验最近日志
def get_recent_logs(experiment_id):
    """一下返回最多 MAX_NUM 条打印记录

    Parameters
    ----------
    experiment_id : int
        实验唯一ID

    Returns
    -------
    dict :
        recent: List[str]
            0: 截止处日志文件
            1: 开始处日志文件
        logs: List[str]
            日志列表
    """

    console_path: str = __get_console_dir_by_id(experiment_id)
    # 是否存在
    if not os.path.exists(console_path):
        return NOT_FOUND_404("Log Folder not Found")
    consoles: list = [f for f in os.listdir(console_path)]
    # 含有error.log，在返回值中带上其中的错误信息
    error = None
    if "error.log" in consoles:
        with open(os.path.join(console_path, "error.log"), mode="r", encoding="utf-8") as f:
            error = f.read().split("\n")
        # 在consoles里删除error.log
        consoles.remove("error.log")
    # 没有日志，也没有错误日志
    if len(consoles) == 0 and error is None:
        return NOT_FOUND_404("No Logs Found")
    # 如果只有错误日志，直接返回
    elif len(consoles) == 0 and error:
        return SUCCESS_200({"error": error})

    total: int = len(consoles)
    # # 如果 total 大于 1, 按照时间排序
    if total > 1:
        consoles = sorted(consoles, key=lambda x: datetime.strptime(x[:-4], "%Y-%m-%d"), reverse=True)
    logs = []
    # # current_page = total
    for index, f in enumerate(consoles, start=1):
        with open(os.path.join(console_path, f), mode="r", encoding="utf-8") as log:
            logs = log.read().split("\n") + logs
            # 如果当前收集到的数据超过限制，退出循环
            # if len(logs) >= MAX_NUM:
            #     # current_page = index
            #     break
    # 如果 logs 内容为空
    if len(logs) == 0:
        return NOT_FOUND_404("No Logs Found")

    # logs = logs[:MAX_NUM]
    end = (logs[-1] if not logs[-1] == "" else logs[-2]).split(" ")[0]
    data = {
        "recent": [logs[0].split(" ")[0], end],
        "logs": logs,
    }
    if error is not None:
        data["error"] = error
    # 返回最新的 MAX_NUM 条记录
    return SUCCESS_200(data)


# 获取图表信息
def get_experimet_charts(experiment_id: int):
    """获取图表信息

    Parameters
    ----------
    experiment_id : int
        实验唯一 ID

    Returns
    -------
    dict :
        _sum: integer
        charts: List[dict]
        namesapces: List[dict]
    """
    chart_list, namespace_list = get_exp_charts(experiment_id)

    return SUCCESS_200(
        {
            "charts": clear_field(chart_list, "experiment_id"),
            "namespaces": namespace_list,
        }
    )


# 更改项目信息
async def update_experiment_info(experiment_id: int, request: Request):
    """修改实验的信息

    Parameters
    ----------
    experiment_id : int
        实验id
    body : Body
        name: str
            实验名称
        description: str
            实验描述

    Returns
    -------
    dict :
        name: str
        description: str
    """

    db = connect()
    body = await request.json()
    body["name"] = body["name"].strip()
    with db.atomic():
        experiment = Experiment.get(experiment_id)
        experiment.name = body.get("name")
        experiment.description = body.get("description")
        experiment.save()

    db.close()

    return SUCCESS_200(body)


# 删除实验
def delete_experiment(experiment_id: int):
    """删除实验

    注意，需要先判断当前实验是否正在运行中，不可删除运行中的实验
    同时，需要删除所有表中含有该 experiment_id 的行

    Parameters
    ----------
    experiment_id : Int
        实验唯一ID

    Returns
    -------
    project : Dictionary
        删除实验后的项目信息，提供给前端更新界面

    TODO: 该怎么删才能高效删干净
    """

    # 先删除实验目录
    experiment_path = __get_exp_dir_by_id(experiment_id)
    shutil.rmtree(experiment_path)
    # 看看该实验下有哪儿些 tag
    tags = Tag.filter(Tag.experiment_id == experiment_id)
    tag_names = [tag["name"] for tag in __to_list(tags)]
    # 检查多实验图表是否有需要删除的
    project_id = Experiment.get_by_id(experiment_id).project_id.id
    # 找到属于多实验且与该实验相关的图表
    charts = Chart.filter(Chart.project_id == project_id, Chart.name.in_(tag_names))

    db = connect()
    with db.atomic():
        # 必须先清除数据库中的实验数据
        Experiment.delete().where(Experiment.id == experiment_id).execute()
        # 图表无 source 的需要删除
        del_list = []
        for chart in charts:
            if chart.sources.count() == 0:
                del_list.append(chart.id)
        Chart.delete().where(Chart.id.in_(del_list)).execute()
    db.commit()

    return SUCCESS_200({"experiment_id": experiment_id})


# 停止实验
def stop_experiment(experiment_id: int):
    """停止实验

    Parameters
    ----------
    experiment_id : Int
        实验唯一ID

    Returns
    -------
    project : Dictionary
        停止实验后的项目信息，提供给前端更新界面
    """
    experiment: Experiment = Experiment.get(experiment_id)
    experiment.update_status(Experiment.STOPPED_STATUS)

    return SUCCESS_200(
        {
            "id": experiment_id,
            "status": Experiment.STOPPED_STATUS,
            "finish_time": create_time(),
        }
    )


# 获取实验依赖
def get_experiment_requirements(experiment_id: int):
    """获取实验依赖

    Parameters
    ----------
    experiment_id : int
        实验唯一ID

    Returns
    -------
    dict :
        requirements: list
            每个依赖项为一行，以列表的形式返回
    """

    path = __get_requirements_path_by_id(experiment_id)
    if not os.path.exists(path):
        return DATA_ERROR_500("failed to find requirements")
    with open(path, encoding="utf-8") as f:
        requirements = f.read()

    return SUCCESS_200({"requirements": requirements.split("\n")})


# 修改实验是否可见
def change_experiment_visibility(experiment_id: int, show: bool):
    """修改实验是否可见

    Parameters
    ----------
    experiment_id : int
        实验id
    show : bool
        在多实验对比图表中，该实验是否可见，true -> 1 , false -> 0

    Returns
    -------
    experiment : dict
        当前实验信息
    """

    try:
        experiment = Experiment.get_by_id(experiment_id)
    except NotExistedError:
        return NOT_FOUND_404("Experiment with id {} does not exist.".format(experiment_id))

    if show:
        experiment.show = 1
    else:
        experiment.show = 0
    experiment.save()
    return SUCCESS_200({"experiment": experiment.__dict__()})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-03-09 21:44:40
@File: swanlab/server/controller/db.py
@IDE: vscode
@Description:
    数据库，从数据库模块中导入数据库相关函数，供其他模块调用，不用每次都通过相对路径导入了
"""
from ..db import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-17 14:56:38
@File: swanlab\db\modules\experiment.py
@IDE: vscode
@Description:
    实验数据表
"""
import os.path

from ..model import SwanModel
from peewee import (
    ForeignKeyField,
    CharField,
    IntegerField,
    TextField,
    IntegrityError,
    Check,
    DatabaseProxy,
)
from .projects import Project
from ..error import ExistedError, NotExistedError, ForeignProNotExistedError
from swanboard.utils import generate_color, get_swanlog_dir, get_package_version
from swankit.env import create_time
import shutil


# 定义模型类
class Experiment(SwanModel):
    """实验表"""

    # 实验运行时状态符
    RUNNING_STATUS = 0
    # 实验停止时状态符
    STOPPED_STATUS = -1
    # 实验结束时状态符
    FINISHED_STATUS = 1

    class Meta:
        database = DatabaseProxy()
        # 通过meta规定name和project_id的唯一性
        indexes = ((("name", "project_id"), True), (("sort", "project_id"), True))
        # sort必须大于等于0
        constraints = [Check("sort >= 0")]

    id = IntegerField(primary_key=True)
    """实验id"""
    project_id = ForeignKeyField(
        Project,
        backref="experiments",
        on_delete="CASCADE",
        on_update="CASCADE",
        default=1,
    )
    """外键，项目id，可通过此外键反向查询项目下的所有实验"""

    run_id = TextField(unique=True)
    """运行时id，用于区分不同的实验保存的文件夹名称"""

    name = CharField(max_length=100, null=False)
    """实验名称，通过meta规定name和project_id的唯一性"""

    description = CharField(max_length=255, null=True)
    """实验描述"""

    sort = IntegerField(null=False)
    """实验索引，用于排序，通过meta规定index和project_id的唯一性"""

    status = IntegerField(choices=[-1, 0, 1], default=0)
    """实验状态，-1: crushed, 0: running, 1: finished，在创建一个实验时，状态默认为0"""

    show = IntegerField(default=1, choices=[0, 1])
    """实验可见性，0: 不可见，1: 可见"""

    light = CharField(max_length=20, null=True)
    """亮色主题颜色"""
    dark = CharField(max_length=20, null=True)
    """暗色主题颜色"""

    pinned_opened = IntegerField(default=1, choices=[0, 1])
    """实验图表置顶部分是否打开，默认值为1，表示打开"""

    hidden_opened = IntegerField(default=0, choices=[0, 1])
    """实验图表隐藏部分是否打开，默认值为0,表示关闭"""

    more = TextField(null=True)
    """更多信息配置，json格式，将在表函数中检查并解析"""

    version = CharField(max_length=30, null=False)
    """创建时的版本号"""

    create_time = CharField(max_length=30, null=False)
    """创建时间"""

    finish_time = CharField(max_length=30, null=True, default=None)
    """实验结束时间"""

    update_time = CharField(max_length=30, null=False)
    """更新的时间"""

    def __dict__(self):
        return {
            "id": self.id,
            "project_id": self.project_id.__dict__(),
            "run_id": self.run_id,
            "name": self.name,
            "description": self.description,
            "sort": self.sort,
            "status": self.status,
            "show": self.show,
            "more": self.more,
            "version": self.version,
            "create_time": self.create_time,
            "update_time": self.update_time,
            "finish_time": self.finish_time,
        }

    @classmethod
    def create(
        cls,
        name: str,
        run_id: str,
        colors: tuple,
        description: str = None,
        project_id: int = Project.DEFAULT_PROJECT_ID,
        num: int = None,
        more: dict = None,
    ) -> "Experiment":
        """覆写继承的create方法，创建一个新的实验

        Parameters
        ----------
        name : str
            实验名称
        run_id : str
            运行时信息
        colors : Tuple[str, str], optional
            实验颜色，[light, dark]，默认为None
        description : str
            实验描述，默认为空字符串
        project_id : int, optional
            关联的项目id，由于目前为单项目模式，所以不需要手动设置此字段，默认置为DEFAULT_PROJECT_ID
        num : int, optional
            历史实验数量，如果不指定则自动计算
        more : dict, optional
            更多配置，在函数内部会转换为字符串格式

        Returns
        -------
        Experiment
            新建的实验实例

        Raises
        -------
        ForeignProNotExistedError
            项目不存在
        ExistedError
            实验已经存在
        """
        # 检查项目是否存在
        if not Project.select().where(Project.id == project_id).exists():
            raise ForeignProNotExistedError("项目不存在")
        # 这个sum是+1以后的值
        _sum = Project.increase_sum(project_id)
        light, dark = colors
        try:
            return super().create(
                name=name,
                run_id=run_id,
                project_id=project_id,
                description=description,
                more=cls.dict2json(more),
                sort=_sum,
                version=get_package_version(),
                light=light,
                dark=dark,
            )
        except IntegrityError:
            raise ExistedError("实验已经存在")

    @classmethod
    def get(cls, *args, **kwargs) -> "Experiment":
        try:
            return super().get(*args, **kwargs)
        except:
            raise NotExistedError("Experiment does not exist: {}".format(kwargs))

    def update_status(self, status: int):
        """更新实验状态

        Parameters
        ----------
        status : int
            实验状态
        """
        self.status = status
        self.finish_time = create_time()
        self.save()

    @classmethod
    def purely_delete(cls, run_id: str):
        """
        删除某一个实验，顺便删除实验的文件夹

        Parameters
        ----------
        run_id : int
            实验run_id
        """
        # 获取实验实例
        try:
            exp = cls.get(run_id=run_id)
            # 删除实验
            exp.delete_instance()
            # 更新项目的实验数量
            Project.decrease_sum()
            # 删除实验文件夹
            shutil.rmtree(os.path.join(get_swanlog_dir(), exp.run_id.__str__()))
        except NotExistedError:
            shutil.rmtree(os.path.join(get_swanlog_dir(), run_id))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-01 15:43:21
@File: swanlab/server/settings.py
@IDE: vscode
@Description:
    服务部分基础配置
"""
import os
import mimetypes
from swanboard.utils import get_swanlog_dir

"""
在此处注册静态文件路径，因为静态文件由vue框架编译后生成，在配置中，编译后的文件存储在/swanlab/template中
入口文件为index.html，网页图标为logo.ico，其他文件为assets文件夹中的文件
"""
# 注册静态文件路径
mimetypes.add_type("application/javascript", ".js")
mimetypes.add_type("text/css", ".css")
# 静态文件路径
FILEPATH = os.path.dirname(os.path.abspath(__file__))
TEMPLATE_PATH = os.path.join(FILEPATH, "template")
ASSETS = os.path.join(TEMPLATE_PATH, "assets")
INDEX = os.path.join(TEMPLATE_PATH, "index.html")
# swanlog文件夹路径
SWANLOG_DIR = get_swanlog_dir()


# ---------------------------------- 实验目录下的各个路径 ----------------------------------

# tags 存储目录
LOGS = "logs"
# files 实验环境目录
FILES = "files"
# console 实验环境目录
CONSOLE = "console"
# media 实验生成媒体目录
MEDIA = "media"
# 实验环境(元数据)
META_DATA = "swanlab-metadata.json"
# 实验依赖
REQUIREMENTS = "requirements.txt"
# 数据库名
DB_NAME = "runs.swanlab"
# 数据库路径
DB_PATH = os.path.join(SWANLOG_DIR, DB_NAME)
# 实验配置
EXPERIMENT_CONFIG = "config.yaml"


def get_exp_dir(name) -> str:
    """实验目录路径"""
    return os.path.join(SWANLOG_DIR, name)


def get_logs_dir(name) -> str:
    """logs 目录路径，存放 tag 数据"""
    return os.path.join(SWANLOG_DIR, name, LOGS)


def get_files_dir(name) -> str:
    """files 目录路径"""
    return os.path.join(SWANLOG_DIR, name, FILES)


def get_console_dir(name) -> str:
    """console 目录路径"""
    return os.path.join(SWANLOG_DIR, name, CONSOLE)


def get_config_path(name) -> str:
    """实验配置存储路径"""
    return os.path.join(get_files_dir(name), EXPERIMENT_CONFIG)


def get_meta_path(name) -> str:
    """实验环境存储路径"""
    return os.path.join(SWANLOG_DIR, name, FILES, META_DATA)


def get_requirements_path(name) -> str:
    """实验依赖存储路径"""
    return os.path.join(SWANLOG_DIR, name, FILES, REQUIREMENTS)


def get_tag_dir(name, tag) -> str:
    """获取 tag 对应的目录路径"""
    return os.path.join(SWANLOG_DIR, name, LOGS, tag)


def get_media_dir(name, tag) -> str:
    """获取 media 对应的目录路径"""
    return os.path.join(SWANLOG_DIR, name, MEDIA, tag)
